import datetime as dt
import json
import os
import re
import sys
import traceback

import pytest
from click.testing import CliRunner
from freezegun import freeze_time

from khal.cli import main_ikhal, main_khal
from khal.utils import CONTENT_ATTRIBUTES

from .utils import _get_ics_filepath, _get_text


class CustomCliRunner(CliRunner):
    def __init__(self, config_file, db=None, calendars=None,
                 xdg_data_home=None, xdg_config_home=None, tmpdir=None, **kwargs) -> None:
        self.config_file = config_file
        self.db = db
        self.calendars = calendars
        self.xdg_data_home = xdg_data_home
        self.xdg_config_home = xdg_config_home
        self.tmpdir = tmpdir

        super().__init__(**kwargs)

    def invoke(self, cli, args=None, *a, **kw):
        args = ['-c', str(self.config_file)] + (args or [])
        return super().invoke(cli, args, *a, **kw)


@pytest.fixture
def runner(tmpdir, monkeypatch):
    db = tmpdir.join('khal.db')
    calendar = tmpdir.mkdir('calendar')
    calendar2 = tmpdir.mkdir('calendar2')
    calendar3 = tmpdir.mkdir('calendar3')

    xdg_data_home = tmpdir.join('vdirs')
    xdg_config_home = tmpdir.join('.config')
    config_file = xdg_config_home.join('khal').join('config')

    # TODO create a vdir config on disk and let vdirsyncer actually read it
    monkeypatch.setattr('vdirsyncer.cli.config.load_config', lambda: Config())
    monkeypatch.setattr('xdg.BaseDirectory.xdg_data_home', str(xdg_data_home))
    monkeypatch.setattr('xdg.BaseDirectory.xdg_config_home', str(xdg_config_home))
    monkeypatch.setattr('xdg.BaseDirectory.xdg_config_dirs', [str(xdg_config_home)])

    def inner(print_new=False, default_calendar=True, days=2, **kwargs):
        if default_calendar:
            default_calendar = 'default_calendar = one'
        else:
            default_calendar = ''
        if not os.path.exists(str(xdg_config_home.join('khal'))):
            os.makedirs(str(xdg_config_home.join('khal')))
        config_file.write(config_template.format(
            delta=str(days) + 'd',
            calpath=str(calendar), calpath2=str(calendar2), calpath3=str(calendar3),
            default_calendar=default_calendar,
            print_new=print_new,
            dbpath=str(db), **kwargs))
        runner = CustomCliRunner(
            config_file=config_file, db=db, calendars={"one": calendar},
            xdg_data_home=xdg_data_home, xdg_config_home=xdg_config_home,
            tmpdir=tmpdir,
        )
        return runner
    return inner


config_template = '''
[calendars]
[[one]]
path = {calpath}
color = dark blue

[[two]]
path = {calpath2}
color = dark green

[[three]]
path = {calpath3}

[locale]
local_timezone = Europe/Berlin
default_timezone = Europe/Berlin

timeformat = %H:%M
dateformat = %d.%m.
longdateformat = %d.%m.%Y
datetimeformat =  %d.%m. %H:%M
longdatetimeformat = %d.%m.%Y %H:%M
firstweekday = 0

[default]
{default_calendar}
timedelta = {delta}
print_new = {print_new}

[sqlite]
path = {dbpath}
'''


def test_direct_modification(runner):
    runner = runner()

    result = runner.invoke(main_khal, ['list'])
    assert result.output == ''
    assert not result.exception

    cal_dt = _get_text('event_dt_simple')
    event = runner.calendars['one'].join('test.ics')
    event.write(cal_dt)
    format = '{start-end-time-style}: {title}'
    args = ['list', '--format', format, '--day-format', '', '09.04.2014']
    result = runner.invoke(main_khal, args)
    assert not result.exception
    assert result.output == '09:30-10:30: An Event\n'

    os.remove(str(event))
    result = runner.invoke(main_khal, ['list'])
    assert not result.exception
    assert result.output == ''


def test_simple(runner):
    runner = runner(days=2)
    result = runner.invoke(main_khal, ['list'])
    assert not result.exception
    assert result.output == ''

    now = dt.datetime.now().strftime('%d.%m.%Y')
    result = runner.invoke(
        main_khal, f'new {now} 18:00 myevent'.split())
    assert result.output == ''
    assert not result.exception

    result = runner.invoke(main_khal, ['list'])
    print(result.output)
    assert 'myevent' in result.output
    assert '18:00' in result.output
    # test show_all_days default value
    assert 'Tomorrow:' not in result.output
    assert not result.exception


def test_simple_color(runner):
    runner = runner(days=2)
    now = dt.datetime.now().strftime('%d.%m.%Y')
    result = runner.invoke(main_khal, f'new {now} 18:00 myevent'.split())
    assert result.output == ''
    assert not result.exception

    result = runner.invoke(main_khal, ['list'], color=True)
    assert not result.exception
    assert '\x1b[34m' in result.output


def test_days(runner):
    runner = runner(days=9)

    when = (dt.datetime.now() + dt.timedelta(days=7)).strftime('%d.%m.%Y')
    result = runner.invoke(main_khal, f'new {when} 18:00 nextweek'.split())
    assert result.output == ''
    assert not result.exception

    when = (dt.datetime.now() + dt.timedelta(days=30)).strftime('%d.%m.%Y')
    result = runner.invoke(main_khal, f'new {when} 18:00 nextmonth'.split())
    assert result.output == ''
    assert not result.exception

    result = runner.invoke(main_khal, ['list'])
    assert 'nextweek' in result.output
    assert 'nextmonth' not in result.output
    assert '18:00' in result.output
    assert not result.exception


def test_notstarted(runner):
    with freeze_time('2015-6-1 15:00'):
        runner = runner(days=2)
        for command in [
                'new 30.5.2015 5.6.2015 long event',
                'new 2.6.2015 4.6.2015 two day event',
                'new 1.6.2015 14:00 18:00 four hour event',
                'new 1.6.2015 16:00 17:00 one hour event',
                'new 2.6.2015 10:00 13:00 three hour event',
        ]:
            result = runner.invoke(main_khal, command.split())
            assert not result.exception

        result = runner.invoke(main_khal, 'list now'.split())
        assert result.output == \
            """Today, 01.06.2015
↔ long event
14:00-18:00 four hour event
16:00-17:00 one hour event
Tomorrow, 02.06.2015
↔ long event
↦ two day event
10:00-13:00 three hour event
Wednesday, 03.06.2015
↔ long event
↔ two day event
"""
        assert not result.exception
        result = runner.invoke(main_khal, 'list now --notstarted'.split())
        assert result.output == \
            """Today, 01.06.2015
16:00-17:00 one hour event
Tomorrow, 02.06.2015
↦ two day event
10:00-13:00 three hour event
Wednesday, 03.06.2015
↔ two day event
"""
        assert not result.exception

        result = runner.invoke(main_khal, 'list now --once'.split())
        assert result.output == \
            """Today, 01.06.2015
↔ long event
14:00-18:00 four hour event
16:00-17:00 one hour event
Tomorrow, 02.06.2015
↦ two day event
10:00-13:00 three hour event
"""
        assert not result.exception

        result = runner.invoke(main_khal, 'list now --once --notstarted'.split())
        assert result.output == \
            """Today, 01.06.2015
16:00-17:00 one hour event
Tomorrow, 02.06.2015
↦ two day event
10:00-13:00 three hour event
"""
        assert not result.exception


def test_calendar(runner):
    with freeze_time('2015-6-1'):
        runner = runner(days=0)
        result = runner.invoke(main_khal, ['calendar'])
        assert not result.exception
        assert result.exit_code == 0
        output = '\n'.join([
            "    Mo Tu We Th Fr Sa Su     No events",
            "Jun  1  2  3  4  5  6  7     ",
            "     8  9 10 11 12 13 14     ",
            "    15 16 17 18 19 20 21     ",
            "    22 23 24 25 26 27 28     ",
            "Jul 29 30  1  2  3  4  5     ",
            "     6  7  8  9 10 11 12     ",
            "    13 14 15 16 17 18 19     ",
            "    20 21 22 23 24 25 26     ",
            "Aug 27 28 29 30 31  1  2     ",
            "     3  4  5  6  7  8  9     ",
            "    10 11 12 13 14 15 16     ",
            "    17 18 19 20 21 22 23     ",
            "    24 25 26 27 28 29 30     ",
            "Sep 31  1  2  3  4  5  6     ",
            "",
        ])
        assert result.output == output


def test_long_calendar(runner):
    with freeze_time('2015-6-1'):
        runner = runner(days=100)
        result = runner.invoke(main_khal, ['calendar'])
        assert not result.exception
        assert result.exit_code == 0
        output = '\n'.join([
            "    Mo Tu We Th Fr Sa Su     No events",
            "Jun  1  2  3  4  5  6  7     ",
            "     8  9 10 11 12 13 14     ",
            "    15 16 17 18 19 20 21     ",
            "    22 23 24 25 26 27 28     ",
            "Jul 29 30  1  2  3  4  5     ",
            "     6  7  8  9 10 11 12     ",
            "    13 14 15 16 17 18 19     ",
            "    20 21 22 23 24 25 26     ",
            "Aug 27 28 29 30 31  1  2     ",
            "     3  4  5  6  7  8  9     ",
            "    10 11 12 13 14 15 16     ",
            "    17 18 19 20 21 22 23     ",
            "    24 25 26 27 28 29 30     ",
            "Sep 31  1  2  3  4  5  6     ",
            "     7  8  9 10 11 12 13     ",
            "    14 15 16 17 18 19 20     ",
            "    21 22 23 24 25 26 27     ",
            "Oct 28 29 30  1  2  3  4     ",
            "",
        ])
        assert result.output == output


def test_default_command_empty(runner):
    runner = runner(days=2)

    result = runner.invoke(main_khal)
    assert result.exception
    assert result.exit_code == 2
    assert result.output.startswith('Usage: ')


def test_invalid_calendar(runner):
    runner = runner(days=2)
    result = runner.invoke(
        main_khal, ['new'] + '-a one 18:00 myevent'.split())
    assert not result.exception
    result = runner.invoke(
        main_khal, ['new'] + '-a inexistent 18:00 myevent'.split())
    assert result.exception
    assert result.exit_code == 2
    assert 'Unknown calendar ' in result.output


def test_attach_calendar(runner):
    runner = runner(days=2)
    result = runner.invoke(main_khal, ['printcalendars'])
    assert set(result.output.split('\n')[:3]) == {'one', 'two', 'three'}
    assert not result.exception
    result = runner.invoke(main_khal, ['printcalendars', '-a', 'one'])
    assert result.output == 'one\n'
    assert not result.exception
    result = runner.invoke(main_khal, ['printcalendars', '-d', 'one'])
    assert set(result.output.split('\n')[:2]) == {'two', 'three'}
    assert not result.exception


# "see #905"
@pytest.mark.xfail
@pytest.mark.parametrize('contents', [
    '',
    'BEGIN:VCALENDAR\nBEGIN:VTODO\nEND:VTODO\nEND:VCALENDAR\n'
])
def test_no_vevent(runner, tmpdir, contents):
    runner = runner(days=2)
    broken_item = runner.calendars['one'].join('broken_item.ics')
    broken_item.write(contents.encode('utf-8'), mode='wb')

    result = runner.invoke(main_khal, ['list'])
    assert not result.exception
    assert result.output == ''


def test_printformats(runner):
    runner = runner(days=2)

    result = runner.invoke(main_khal, ['printformats'])
    assert '\n'.join(['longdatetimeformat: 21.12.2013 21:45',
                      'datetimeformat: 21.12. 21:45',
                      'longdateformat: 21.12.2013',
                      'dateformat: 21.12.',
                      'timeformat: 21:45',
                      '']) == result.output
    assert not result.exception


# "see #810"
@pytest.mark.xfail
def test_repeating(runner):
    runner = runner(days=2)
    now = dt.datetime.now().strftime('%d.%m.%Y')
    end_date = dt.datetime.now() + dt.timedelta(days=10)
    result = runner.invoke(
        main_khal, (f"new {now} 18:00 myevent -r weekly -u "
                    f"{end_date.strftime('%d.%m.%Y')}").split())
    assert not result.exception
    assert result.output == ''


def test_at(runner):
    runner = runner(days=2)
    now = dt.datetime.now().strftime('%d.%m.%Y')
    end_date = dt.datetime.now() + dt.timedelta(days=10)
    result = runner.invoke(
        main_khal,
        f"new {now} {end_date.strftime('%d.%m.%Y')} 18:00 myevent".split())
    args = ['--color', 'at', '--format', '{start-time}{title}', '--day-format', '', '18:30']
    result = runner.invoke(main_khal, args)
    assert not result.exception
    assert result.output.startswith('myevent')


def test_at_json(runner):
    runner = runner(days=2)
    now = dt.datetime.now().strftime('%d.%m.%Y')
    end_date = dt.datetime.now() + dt.timedelta(days=10)
    result = runner.invoke(
        main_khal,
        'new {} {} 18:00 myevent'.format(now, end_date.strftime('%d.%m.%Y')).split())
    args = ['--color', 'at', '--json', 'start-time', '--json', 'title', '18:30']
    result = runner.invoke(main_khal, args)
    assert not result.exception
    assert result.output.startswith('[{"start-time": "", "title": "myevent"}]')


def test_at_json_default_fields(runner):
    runner = runner(days=2)
    now = dt.datetime.now().strftime('%d.%m.%Y')
    end_date = dt.datetime.now() + dt.timedelta(days=10)
    result = runner.invoke(
        main_khal,
        'new {} {} 18:00 myevent'.format(now, end_date.strftime('%d.%m.%Y')).split())
    args = ['--color', 'at', '--json', 'all', '18:30']
    result = runner.invoke(main_khal, args)
    assert not result.exception
    output_fields = json.loads(result.output)[0].keys()
    assert all(x in output_fields for x in CONTENT_ATTRIBUTES)


def test_at_json_strip(runner):
    runner = runner()
    result = runner.invoke(main_khal, ['import', _get_ics_filepath(
        'event_rrule_recuid_cancelled')], input='0\ny\n')
    assert not result.exception
    result = runner.invoke(main_khal, ['at', '--json', 'repeat-symbol',
                           '--json', 'status', '--json', 'cancelled', '14.07.2014', '07:00'])
    traceback.print_tb(result.exc_info[2])
    assert not result.exception
    assert result.output.startswith(
        '[{"repeat-symbol": "⟳", "status": "CANCELLED", "cancelled": "CANCELLED"}]')


def test_at_day_format(runner):
    runner = runner(days=2)
    now = dt.datetime.now().strftime('%d.%m.%Y')
    end_date = dt.datetime.now() + dt.timedelta(days=10)
    result = runner.invoke(
        main_khal,
        f"new {now} {end_date.strftime('%d.%m.%Y')} 18:00 myevent".split())
    args = ['--color', 'at', '--format', '{start-time}{title}', '--day-format', '{name}', '18:30']
    result = runner.invoke(main_khal, args)
    assert not result.exception
    assert result.output.startswith('Today\x1b[0m\nmyevent')


def test_list(runner):
    runner = runner(days=2)
    now = dt.datetime.now().strftime('%d.%m.%Y')
    result = runner.invoke(
        main_khal,
        f'new {now} 18:00 myevent'.split())
    format = '{red}{start-end-time-style}{reset} {title} :: {description}'
    args = ['--color', 'list', '--format', format, '--day-format', 'header', '18:30']
    result = runner.invoke(main_khal, args)
    expected = 'header\x1b[0m\n\x1b[31m18:00-19:00\x1b[0m myevent :: \x1b[0m\n'
    assert not result.exception
    assert result.output.startswith(expected)


def test_list_json(runner):
    runner = runner(days=2)
    now = dt.datetime.now().strftime('%d.%m.%Y')
    result = runner.invoke(
        main_khal,
        f'new {now} 18:00 myevent'.split())
    args = ['list', '--json', 'start-end-time-style',
            '--json', 'title', '--json', 'description', '18:30']
    result = runner.invoke(main_khal, args)
    expected = '[{"start-end-time-style": "18:00-19:00", "title": "myevent", "description": ""}]'
    assert not result.exception
    assert result.output.startswith(expected)


def test_search(runner):
    runner = runner(days=2)
    now = dt.datetime.now().strftime('%d.%m.%Y')
    result = runner.invoke(main_khal, f'new {now} 18:00 myevent'.split())
    format = '{red}{start-end-time-style}{reset} {title} :: {description}'
    result = runner.invoke(main_khal, ['--color', 'search', '--format', format, 'myevent'])
    assert not result.exception
    assert result.output.startswith('\x1b[34m\x1b[31m18:00')


def test_search_json(runner):
    runner = runner(days=2)
    now = dt.datetime.now().strftime('%d.%m.%Y')
    result = runner.invoke(main_khal, f'new {now} 18:00 myevent'.split())
    result = runner.invoke(main_khal, ['search', '--json', 'start-end-time-style',
                                       '--json', 'title', '--json', 'description', 'myevent'])
    assert not result.exception
    assert result.output.startswith('[{"start-end-time-style": "18:00')


def test_no_default_new(runner):
    runner = runner(default_calendar=False)
    result = runner.invoke(main_khal, 'new 18:00 beer'.split())
    assert ("Error: Invalid value: No default calendar is configured, "
            "please provide one explicitly.") in result.output
    assert result.exit_code == 2

def test_print_bad_ics(runner):
    """Attempt to print a .ics that is malformed, but does not have a DST-related error."""
    runner = runner()
    result = runner.invoke(main_khal, ['printics', _get_ics_filepath('non_dst_error')])
    assert result.exception
    expected = ValueError("Invalid iCalendar duration: PT-2H")
    assert expected.__class__ == result.exception.__class__
    assert expected.args == result.exception.args

def test_import(runner, monkeypatch):
    runner = runner()
    result = runner.invoke(main_khal, 'import -a one -a two import file.ics'.split())
    assert result.exception
    assert result.exit_code == 2
    assert 'Can\'t use "--include-calendar" / "-a" more than once' in result.output

    class FakeImport:
        args, kwargs = None, None

        def clean(self):
            self.args, self.kwargs = None, None

        def import_ics(self, *args, **kwargs):
            print('saving args')
            print(args)
            self.args = args
            self.kwargs = kwargs

    fake = FakeImport()
    monkeypatch.setattr('khal.controllers.import_ics', fake.import_ics)
    # as we are not actually parsing the file we want to import, we can use
    # any readable file at all, therefore re-using the configuration file
    result = runner.invoke(main_khal, f'import -a one {runner.config_file}'.split())
    assert not result.exception
    assert {cal['name'] for cal in fake.args[0].calendars} == {'one'}

    fake.clean()
    result = runner.invoke(main_khal, f'import {runner.config_file}'.split())
    assert not result.exception
    assert {cal['name'] for cal in fake.args[0].calendars} == {'one', 'two', 'three'}


def test_import_proper(runner):
    runner = runner()
    result = runner.invoke(main_khal, ['import', _get_ics_filepath('cal_d')], input='0\ny\n')
    assert result.output.startswith('09.04.-09.04. An Event')
    assert not result.exception
    result = runner.invoke(main_khal, ['search', 'Event'])
    assert result.output == '09.04.-09.04. An Event\n'


def test_import_proper_invalid_timezone(runner):
    runner = runner()
    result = runner.invoke(
        main_khal, ['import', _get_ics_filepath('invalid_tzoffset')], input='0\ny\n')
    assert result.output.startswith(
        'warning: Invalid timezone offset encountered, timezone information may be wrong')
    assert not result.exception
    result = runner.invoke(main_khal, ['search', 'Event'])
    assert result.output.startswith(
        'warning: Invalid timezone offset encountered, timezone information may be wrong')
    assert '02.12. 08:00-02.12. 09:30 Some event' in result.output


def test_import_invalid_choice_and_prefix(runner):
    runner = runner()
    result = runner.invoke(main_khal, ['import', _get_ics_filepath('cal_d')], input='9\nth\ny\n')
    assert result.output.startswith('09.04.-09.04. An Event')
    assert result.output.find('invalid choice') == 125
    assert not result.exception
    result = runner.invoke(main_khal, ['search', 'Event'])
    assert result.output == '09.04.-09.04. An Event\n'


def test_import_from_stdin(runner, monkeypatch):
    ics_data = 'This is some really fake icalendar data'

    class FakeImport:
        args, kwargs = None, None
        call_count = 0

        def clean(self):
            self.args, self.kwargs = None, None

        def import_ics(self, *args, **kwargs):
            print('saving args')
            print(args)
            self.call_count += 1
            self.args = args
            self.kwargs = kwargs

    importer = FakeImport()
    monkeypatch.setattr('khal.controllers.import_ics', importer.import_ics)
    runner = runner()
    result = runner.invoke(main_khal, ['import'], input=ics_data)

    assert not result.exception
    assert importer.call_count == 1
    assert importer.kwargs['ics'] == ics_data


def test_interactive_command(runner, monkeypatch):
    runner = runner(days=2)
    token = "hooray"

    def fake_ui(*a, **kw):
        print(token)
        sys.exit(0)

    monkeypatch.setattr('khal.ui.start_pane', fake_ui)

    result = runner.invoke(main_ikhal, ['-a', 'one'])
    assert not result.exception
    assert result.output.strip() == token

    result = runner.invoke(main_khal, ['interactive', '-a', 'one'])
    assert not result.exception
    assert result.output.strip() == token


def test_color_option(runner):
    runner = runner(days=2)

    result = runner.invoke(main_khal, ['--no-color', 'list'])
    assert result.output == ''

    result = runner.invoke(main_khal, ['--color', 'list'])
    assert result.output == ''


def choices(dateformat=0, timeformat=0,
            default_calendar='',
            calendar_option=0,
            accept_vdirsyncer_dir=True,
            vdir='',
            caldav_url='', caldav_user='', caldav_pw='',
            write_config=True):
    """helper function to generate input for testing `configure`"""
    confirm = {True: 'y', False: 'n'}

    out = [
        str(dateformat), str(timeformat), str(calendar_option),
    ]
    if calendar_option == 1:
        out.append(confirm[accept_vdirsyncer_dir])
        if not accept_vdirsyncer_dir:
            out.append(vdir)
    elif calendar_option == 2:
        out.append(caldav_url)
        out.append(caldav_user)
        out.append(caldav_pw)
    out.append(default_calendar)
    out.append(confirm[write_config])
    out.append('')
    return '\n'.join(out)


class Config:
    """helper class for mocking vdirsyncer's config objects"""
    # TODO crate a vdir config on disk and let vdirsyncer actually read it
    storages = {
        'home_calendar_local': {
            'type': 'filesystem',
            'instance_name': 'home_calendar_local',
            'path': '~/.local/share/calendars/home/',
            'fileext': '.ics',
        },
        'events_local': {
            'type': 'filesystem',
            'instance_name': 'events_local',
            'path': '~/.local/share/calendars/events/',
            'fileext': '.ics',
        },
        'home_calendar_remote': {
            'type': 'caldav',
            'url': 'https://some.url/caldav',
            'username': 'foo',
            'password.fetch': ['command', 'get_secret'],
            'instance_name': 'home_calendar_remote',
        },
        'home_contacts_remote': {
            'type': 'carddav',
            'url': 'https://another.url/caldav',
            'username': 'bar',
            'password.fetch': ['command', 'get_secret'],
            'instance_name': 'home_contacts_remote',
        },
        'home_contacts_local': {
            'type': 'filesystem',
            'instance_name': 'home_contacts_local',
            'path': '~/.local/share/contacts/',
            'fileext': '.vcf',
        },
        'events_remote': {
            'type': 'http',
            'instance_name': 'events_remote',
            'url': 'http://list.of/events/',
        },
    }


def test_configure_command(runner):
    runner_factory = runner
    runner = runner()
    runner.config_file.remove()

    result = runner.invoke(main_khal, ['configure'], input=choices())
    assert f'Successfully wrote configuration to {runner.config_file}' in result.output
    assert result.exit_code == 0
    with open(str(runner.config_file)) as f:
        actual_config = ''.join(f.readlines())

    assert actual_config == f'''[calendars]

[[private]]
path = {runner.tmpdir}/vdirs/khal/calendars/private
type = calendar

[locale]
timeformat = %H:%M
dateformat = %Y-%m-%d
longdateformat = %Y-%m-%d
datetimeformat = %Y-%m-%d %H:%M
longdatetimeformat = %Y-%m-%d %H:%M

[default]
default_calendar = private
'''

    # if aborting, no config file should be written
    runner = runner_factory()
    assert os.path.exists(str(runner.config_file))
    runner.config_file.remove()
    assert not os.path.exists(str(runner.config_file))

    result = runner.invoke(main_khal, ['configure'], input=choices(write_config=False))
    assert 'aborted' in result.output
    assert result.exit_code == 1


def test_print_ics_command(runner):
    runner = runner()
    # Input is empty and loading from stdin
    result = runner.invoke(main_khal, ['printics', '-'])
    assert result.exception

    # Non existing file
    result = runner.invoke(main_khal, ['printics', 'nonexisting_file'])
    assert result.exception
    assert re.search(r'''Error: Invalid value for "?'?\[?(ICS|ics)\]?'?"?: '''
                     r'''('nonexisting_file': No such file or directory\n|'''
                     r'Could not open file:)', result.output)

    # Run on test files
    result = runner.invoke(main_khal, ['printics', _get_ics_filepath('cal_d')])
    assert not result.exception
    result = runner.invoke(main_khal, ['printics', _get_ics_filepath('cal_dt_two_tz')])
    assert not result.exception

    # Test with some nice format strings
    form = '{uid}\t{title}\t{description}\t{start}\t{start-long}\t{start-date}' \
           '\t{start-date-long}\t{start-time}\t{end}\t{end-long}\t{end-date}' \
           '\t{end-date-long}\t{end-time}\t{repeat-symbol}\t{description}' \
           '\t{description-separator}\t{location}\t{calendar}' \
           '\t{calendar-color}\t{start-style}\t{to-style}\t{end-style}' \
           '\t{start-end-time-style}\t{end-necessary}\t{end-necessary-long}'
    result = runner.invoke(main_khal, [
        'printics', '-f', form, _get_ics_filepath('cal_dt_two_tz')])
    assert not result.exception
    assert 25 == len(result.output.split('\t'))
    result = runner.invoke(main_khal, [
        'printics', '-f', form, _get_ics_filepath('cal_dt_two_tz')])
    assert not result.exception
    assert 25 == len(result.output.split('\t'))


def test_printics_read_from_stdin(runner):
    runner = runner(command='printics')
    result = runner.invoke(main_khal, ['printics'], input=_get_text('cal_d'))
    assert not result.exception
    assert '1 events found in stdin input\n09.04.-09.04. An Event\n' in result.output


def test_configure_command_config_exists(runner):
    runner = runner()
    result = runner.invoke(main_khal, ['configure'], input=choices())
    assert 'Found an existing' in result.output
    assert result.exit_code == 1


def test_configure_command_create_vdir(runner):
    runner = runner()
    runner.config_file.remove()
    runner.xdg_config_home.remove()

    result = runner.invoke(
        main_khal, ['configure'],
        input=choices(),
    )
    assert f'Successfully wrote configuration to {str(runner.config_file)}' in result.output
    assert result.exit_code == 0
    with open(str(runner.config_file)) as f:
        actual_config = ''.join(f.readlines())

    assert actual_config == f'''[calendars]

[[private]]
path = {str(runner.xdg_data_home)}/khal/calendars/private
type = calendar

[locale]
timeformat = %H:%M
dateformat = %Y-%m-%d
longdateformat = %Y-%m-%d
datetimeformat = %Y-%m-%d %H:%M
longdatetimeformat = %Y-%m-%d %H:%M

[default]
default_calendar = private
'''

    # running configure again, should yield another vdir path, as the old
    # one still exists
    runner.config_file.remove()
    result = runner.invoke(
        main_khal, ['configure'],
        input=choices(),
    )
    assert f'Successfully wrote configuration to {str(runner.config_file)}' in result.output
    assert result.exit_code == 0
    with open(str(runner.config_file)) as f:
        actual_config = ''.join(f.readlines())

    assert f'{runner.xdg_data_home}/khal/calendars/private1' in actual_config


def cleanup(paths):
    """reset permissions of all files and folders in `paths` to 644 resp. 755"""
    for path in paths:
        if os.path.exists(path):
            os.chmod(str(path), 0o755)
        for dirpath, _dirnames, filenames in os.walk(path):
            os.chmod(str(dirpath), 0o755)
            for filename in filenames:
                os.chmod(str(os.path.join(dirpath, filename)), 0o644)


def test_configure_command_cannot_write_config_file(runner):
    runner = runner()
    runner.config_file.remove()
    os.chmod(str(runner.xdg_config_home), 555)
    result = runner.invoke(main_khal, ['configure'], input=choices())
    assert result.exit_code == 1
    # make sure pytest can clean up behind us
    cleanup([runner.xdg_config_home])


def test_configure_command_cannot_create_vdir(runner):
    runner = runner()
    runner.config_file.remove()
    os.mkdir(str(runner.xdg_data_home), mode=555)
    result = runner.invoke(
        main_khal, ['configure'],
        input=choices(),
    )
    assert 'Exiting' in result.output
    assert result.exit_code == 1
    # make sure pytest can clean up behind us
    cleanup([runner.xdg_data_home])


def test_edit(runner):
    runner = runner()
    result = runner.invoke(main_khal, ['list'])
    assert not result.exception
    assert result.output == ''

    for name in ['event_dt_simple', 'event_d_15']:
        cal_dt = _get_text(name)
        event = runner.calendars['one'].join(f'{name}.ics')
        event.write(cal_dt)

    format = '{start-end-time-style}: {title}'
    result = runner.invoke(
        main_khal, ['edit', '--show-past', 'Event'], input='s\nGreat Event\nn\nn\n')
    assert not result.exception

    args = ['list', '--format', format, '--day-format', '', '09.04.2014']
    result = runner.invoke(main_khal, args)
    assert '09:30-10:30: Great Event' in result.output
    assert not result.exception

    args = ['list', '--format', format, '--day-format', '', '09.04.2015']
    result = runner.invoke(main_khal, args)
    assert ': An Event' in result.output
    assert not result.exception


def test_new(runner):
    runner = runner(print_new='path')

    result = runner.invoke(main_khal, 'new 13.03.2016 3d Visit'.split())
    assert not result.exception
    assert result.output.endswith('.ics\n')
    assert result.output.startswith(str(runner.tmpdir))


def test_new_format(runner):
    runner = runner(print_new='event')

    format = '{start-end-time-style}: {title}'
    result = runner.invoke(main_khal, ['new', '13.03.2016 12:00', '3d',
                           '--format', format, 'Visit'])
    assert not result.exception
    assert result.output.startswith('→12:00: Visit')


def test_new_json(runner):
    runner = runner(print_new='event')

    result = runner.invoke(main_khal, ['new', '13.03.2016 12:00', '3d',
                           '--json', 'start-end-time-style', '--json', 'title', 'Visit'])
    assert not result.exception
    assert result.output.startswith(
        '[{"start-end-time-style": "→12:00", "title": "Visit"}]')


@ freeze_time('2015-6-1 8:00')
def test_new_interactive(runner):
    runner = runner(print_new='path')

    result = runner.invoke(
        main_khal, 'new -i'.split(),
        'Another event\n13:00 17:00\n\nNone\nn\n'
    )
    assert not result.exception
    assert result.exit_code == 0


def test_debug(runner):
    runner = runner()
    result = runner.invoke(main_khal, ['-v', 'debug', 'printformats'])
    assert result.output.startswith('debug: khal 0.')
    assert 'using the config file at' in result.output
    assert 'debug: Using config:\ndebug: [calendars]' in result.output
    assert not result.exception


@freeze_time('2015-6-1 8:00')
def test_new_interactive_extensive(runner):
    runner = runner(print_new='path', default_calendar=False)

    result = runner.invoke(
        main_khal, 'new -i 15:00 15:30'.split(),
        '?\ninvalid\ntwo\n'
        'Unicce Name\n'
        '\n'
        'Europe/London\n'
        'bar\n'
        'l\non a boat\n'
        'p\nweekly\n'
        '1.1.2018\n'
        'a\n30m\n'
        'c\nwork\n'
        'n\n'
    )
    assert not result.exception
    assert result.exit_code == 0


@freeze_time('2015-6-1 8:00')
def test_issue_1056(runner):
    """if an ansi escape sequence is contained in the output, we can't parse it
    properly"""

    runner = runner(print_new='path', default_calendar=False)

    result = runner.invoke(
        main_khal, 'new -i'.split(),
        'two\n'
        'new event\n'
        'now\n'
        'Europe/London\n'
        'None\n'
        't\n'  # edit datetime range
        '\n'
        'n\n'
    )
    assert 'error parsing range' not in result.output
    assert not result.exception
    assert result.exit_code == 0


def test_list_now(runner, tmpdir):
    # reproduce #693
    runner = runner()

    xdg_config_home = tmpdir.join('.config')
    config_file = xdg_config_home.join('khal').join('config')
    config_file.write("""
        [calendars]
        [[one]]
        path = {}
        color = dark blue
        [[two]]
        path = {}
        color = dark green
        [[three]]
        path = {}
        [locale]
        longdateformat = %a %Y-%m-%d
        dateformat = %Y-%m-%d
    """.format(
        tmpdir.join('calendar'),
        tmpdir.join('calendar2'),
        tmpdir.join('calendar3'),
    ))

    result = runner.invoke(main_khal, ['list', 'now'])
    assert not result.exception



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
from time import sleep

from khal.khalendar import vdir


def test_etag(tmpdir, sleep_time):
    fpath = os.path.join(str(tmpdir), 'foo')

    file_ = open(fpath, 'w')
    file_.write('foo')
    file_.close()

    old_etag = vdir.get_etag_from_file(fpath)
    sleep(sleep_time)

    file_ = open(fpath, 'w')
    file_.write('foo')
    file_.close()

    new_etag = vdir.get_etag_from_file(fpath)

    assert old_etag != new_etag


def test_etag_sync(tmpdir, sleep_time):
    fpath = os.path.join(str(tmpdir), 'foo')

    file_ = open(fpath, 'w')
    file_.write('foo')
    file_.close()
    os.sync()
    old_etag = vdir.get_etag_from_file(fpath)
    sleep(sleep_time)

    file_ = open(fpath, 'w')
    file_.write('foo')
    file_.close()

    new_etag = vdir.get_etag_from_file(fpath)

    assert old_etag != new_etag


def test_get_href_from_uid():
    # Test UID with unsafe characters
    uid = "V042MJ8B3SJNFXQOJL6P53OFMHJE8Z3VZWÈÉ@pimutils.org"
    first_href = vdir._generate_href(uid)
    second_href = vdir._generate_href(uid)
    assert first_href == second_href

    # test UID with safe characters
    uid = "V042MJ8B3SJNFXQOJL6P53OFMHJE8Z3VZWOU@pimutils.org"
    href = vdir._generate_href(uid)
    assert href == uid

    href = vdir._generate_href()
    assert href is not None



import datetime as dt
from operator import itemgetter

import icalendar
import pkg_resources
import pytest

from khal.khalendar import backend
from khal.khalendar.exceptions import OutdatedDbVersionError, UpdateFailed

from .utils import BERLIN, LOCALE_BERLIN, _get_text

calname = 'home'


def test_new_db_version():
    dbi = backend.SQLiteDb(calname, ':memory:', locale=LOCALE_BERLIN)
    backend.DB_VERSION += 1
    with pytest.raises(OutdatedDbVersionError):
        dbi._check_table_version()


def test_event_rrule_recurrence_id():
    dbi = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    assert dbi.list(calname) == []
    events = dbi.get_localized(
        BERLIN.localize(dt.datetime(2014, 6, 30, 0, 0)),
        BERLIN.localize(dt.datetime(2014, 8, 26, 0, 0)),
    )
    assert list(events) == []
    dbi.update(_get_text('event_rrule_recuid'), href='12345.ics', etag='abcd', calendar=calname)
    assert dbi.list(calname) == [('12345.ics', 'abcd')]
    events = dbi.get_localized(
        BERLIN.localize(dt.datetime(2014, 6, 30, 0, 0)),
        BERLIN.localize(dt.datetime(2014, 8, 26, 0, 0)),
    )
    events = sorted(events, key=itemgetter(2))
    assert len(events) == 6

    # start
    assert events[0][2] == BERLIN.localize(dt.datetime(2014, 6, 30, 7, 0))
    assert events[1][2] == BERLIN.localize(dt.datetime(2014, 7, 7, 9, 0))
    assert events[2][2] == BERLIN.localize(dt.datetime(2014, 7, 14, 7, 0))
    assert events[3][2] == BERLIN.localize(dt.datetime(2014, 7, 21, 7, 0))
    assert events[4][2] == BERLIN.localize(dt.datetime(2014, 7, 28, 7, 0))
    assert events[5][2] == BERLIN.localize(dt.datetime(2014, 8, 4, 7, 0))

    calendars = dbi.get_localized_calendars(
        BERLIN.localize(dt.datetime(2014, 6, 30, 0, 0)),
        BERLIN.localize(dt.datetime(2014, 8, 26, 0, 0)),
    )
    calendars = list(calendars)
    assert len(calendars) == 6


def test_event_rrule_recurrence_id_invalid_tzid():
    dbi = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    dbi.update(_get_text('event_rrule_recuid_invalid_tzid'), href='12345.ics', etag='abcd',
               calendar=calname)
    events = dbi.get_localized(
        BERLIN.localize(dt.datetime(2014, 4, 30, 0, 0)),
        BERLIN.localize(dt.datetime(2014, 9, 26, 0, 0)))
    events = sorted(events, key=itemgetter(2))
    assert len(events) == 6

    assert events[0][2] == BERLIN.localize(dt.datetime(2014, 6, 30, 7, 0))
    assert events[1][2] == BERLIN.localize(dt.datetime(2014, 7, 7, 9, 0))
    assert events[2][2] == BERLIN.localize(dt.datetime(2014, 7, 14, 7, 0))
    assert events[3][2] == BERLIN.localize(dt.datetime(2014, 7, 21, 7, 0))
    assert events[4][2] == BERLIN.localize(dt.datetime(2014, 7, 28, 7, 0))
    assert events[5][2] == BERLIN.localize(dt.datetime(2014, 8, 4, 7, 0))


event_rrule_recurrence_id_reverse = """
BEGIN:VCALENDAR
BEGIN:VEVENT
UID:event_rrule_recurrence_id
SUMMARY:Arbeit
RECURRENCE-ID:20140707T050000Z
DTSTART;TZID=Europe/Berlin:20140707T090000
DTEND;TZID=Europe/Berlin:20140707T140000
END:VEVENT
BEGIN:VEVENT
UID:event_rrule_recurrence_id
SUMMARY:Arbeit
RRULE:FREQ=WEEKLY;COUNT=6
DTSTART;TZID=Europe/Berlin:20140630T070000
DTEND;TZID=Europe/Berlin:20140630T120000
END:VEVENT
END:VCALENDAR
"""


def test_event_rrule_recurrence_id_reverse():
    """as icalendar elements can be saved in arbitrary order, we also have to
    deal with `reverse` ordered icalendar files
    """
    dbi = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    assert dbi.list(calname) == []
    events = dbi.get_localized(
        BERLIN.localize(dt.datetime(2014, 6, 30, 0, 0)),
        BERLIN.localize(dt.datetime(2014, 8, 26, 0, 0)))
    assert list(events) == []
    dbi.update(event_rrule_recurrence_id_reverse, href='12345.ics', etag='abcd', calendar=calname)
    assert dbi.list(calname) == [('12345.ics', 'abcd')]
    events = dbi.get_localized(
        BERLIN.localize(dt.datetime(2014, 6, 30, 0, 0)),
        BERLIN.localize(dt.datetime(2014, 8, 26, 0, 0)))
    events = sorted(events, key=itemgetter(2))
    assert len(events) == 6

    assert events[0][2] == BERLIN.localize(dt.datetime(2014, 6, 30, 7, 0))
    assert events[1][2] == BERLIN.localize(dt.datetime(2014, 7, 7, 9, 0))
    assert events[2][2] == BERLIN.localize(dt.datetime(2014, 7, 14, 7, 0))
    assert events[3][2] == BERLIN.localize(dt.datetime(2014, 7, 21, 7, 0))
    assert events[4][2] == BERLIN.localize(dt.datetime(2014, 7, 28, 7, 0))
    assert events[5][2] == BERLIN.localize(dt.datetime(2014, 8, 4, 7, 0))


def test_event_rrule_recurrence_id_update_with_exclude():
    """
    test if updates work as they should. The updated event has the extra
    RECURRENCE-ID event removed and one recurrence date excluded via EXDATE
    """
    dbi = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    dbi.update(_get_text('event_rrule_recuid'), href='12345.ics', etag='abcd', calendar=calname)
    dbi.update(_get_text('event_rrule_recuid_update'),
               href='12345.ics', etag='abcd', calendar=calname)
    events = dbi.get_localized(BERLIN.localize(dt.datetime(2014, 4, 30, 0, 0)),
                               BERLIN.localize(dt.datetime(2014, 9, 26, 0, 0)))
    events = sorted(events, key=itemgetter(2))
    assert len(events) == 5
    assert events[0][2] == BERLIN.localize(dt.datetime(2014, 6, 30, 7, 0))
    assert events[1][2] == BERLIN.localize(dt.datetime(2014, 7, 7, 7, 0))
    assert events[2][2] == BERLIN.localize(dt.datetime(2014, 7, 21, 7, 0))
    assert events[3][2] == BERLIN.localize(dt.datetime(2014, 7, 28, 7, 0))
    assert events[4][2] == BERLIN.localize(dt.datetime(2014, 8, 4, 7, 0))


def test_event_recuid_no_master():
    """
    test for events which have a RECUID component, but the master event is
    not present in the same file
    """
    dbi = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    dbi.update(_get_text('event_dt_recuid_no_master'),
               href='12345.ics', etag='abcd', calendar=calname)
    events = dbi.get_floating(
        dt.datetime(2017, 3, 1, 0, 0), dt.datetime(2017, 4, 1, 0, 0),
    )
    events = sorted(events, key=itemgetter(2))
    assert len(events) == 1
    assert events[0][2] == dt.datetime(2017, 3, 29, 16)
    assert events[0][3] == dt.datetime(2017, 3, 29, 16, 25)
    assert 'SUMMARY:Infrastructure Planning' in events[0][0]


def test_event_recuid_rrule_no_master():
    """
    test for events which have a RECUID and a RRULE component, but the master event is
    not present in the same file
    """
    dbi = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    dbi.update(
        _get_text('event_dt_multi_recuid_no_master'),
        href='12345.ics', etag='abcd', calendar=calname,
    )
    events = dbi.get_floating(
        dt.datetime(2010, 1, 1, 0, 0), dt.datetime(2020, 1, 1, 0, 0),
    )
    events = sorted(events, key=itemgetter(2))
    assert len(list(events)) == 2
    assert events[0][2] == dt.datetime(2014, 6, 30, 7, 30)
    assert events[0][3] == dt.datetime(2014, 6, 30, 12, 0)
    assert events[1][2] == dt.datetime(2014, 7, 7, 8, 30)
    assert events[1][3] == dt.datetime(2014, 7, 7, 12, 0)
    events = dbi.search('VEVENT')
    assert len(list(events)) == 2


def test_no_valid_timezone():
    dbi = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    dbi.update(_get_text('event_dt_local_missing_tz'),
               href='12345.ics', etag='abcd', calendar=calname)
    events = dbi.get_localized(BERLIN.localize(dt.datetime(2014, 4, 9, 0, 0)),
                               BERLIN.localize(dt.datetime(2014, 4, 10, 0, 0)))
    events = sorted(events)
    assert len(events) == 1
    event = events[0]
    assert event[2] == BERLIN.localize(dt.datetime(2014, 4, 9, 9, 30))
    assert event[3] == BERLIN.localize(dt.datetime(2014, 4, 9, 10, 30))


def test_event_delete():
    dbi = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    assert dbi.list(calname) == []
    events = dbi.get_localized(BERLIN.localize(dt.datetime(2014, 6, 30, 0, 0)),
                               BERLIN.localize(dt.datetime(2014, 8, 26, 0, 0)))
    assert list(events) == []
    dbi.update(event_rrule_recurrence_id_reverse, href='12345.ics', etag='abcd', calendar=calname)
    assert dbi.list(calname) == [('12345.ics', 'abcd')]
    events = dbi.get_localized(BERLIN.localize(dt.datetime(2014, 6, 30, 0, 0)),
                               BERLIN.localize(dt.datetime(2014, 9, 26, 0, 0)))
    assert len(list(events)) == 6
    dbi.delete('12345.ics', calendar=calname)
    events = dbi.get_localized(BERLIN.localize(dt.datetime(2014, 6, 30, 0, 0)),
                               BERLIN.localize(dt.datetime(2014, 9, 26, 0, 0)))
    assert len(list(events)) == 0


event_rrule_this_and_prior = """
BEGIN:VCALENDAR
BEGIN:VEVENT
UID:event_rrule_recurrence_id_this_and_prior
SUMMARY:Arbeit
RRULE:FREQ=WEEKLY;UNTIL=20140806T060000Z
DTSTART;TZID=Europe/Berlin:20140630T070000
DTEND;TZID=Europe/Berlin:20140630T120000
END:VEVENT
BEGIN:VEVENT
UID:event_rrule_recurrence_id_this_and_prior
SUMMARY:Arbeit
RECURRENCE-ID;RANGE=THISANDPRIOR:20140707T050000Z
DTSTART;TZID=Europe/Berlin:20140707T090000
DTEND;TZID=Europe/Berlin:20140707T140000
END:VEVENT
END:VCALENDAR
"""


def test_this_and_prior():
    """we do not support THISANDPRIOR, therefore this should fail"""
    dbi = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    with pytest.raises(UpdateFailed):
        dbi.update(event_rrule_this_and_prior, href='12345.ics', etag='abcd', calendar=calname)


event_rrule_this_and_future_temp = """
BEGIN:VCALENDAR
BEGIN:VEVENT
UID:event_rrule_recurrence_id
SUMMARY:Arbeit
RRULE:FREQ=WEEKLY;UNTIL=20140806T060000Z
DTSTART;TZID=Europe/Berlin:20140630T070000
DTEND;TZID=Europe/Berlin:20140630T120000
END:VEVENT
BEGIN:VEVENT
UID:event_rrule_recurrence_id
SUMMARY:Arbeit (lang)
RECURRENCE-ID;RANGE=THISANDFUTURE:20140707T050000Z
DTSTART;TZID=Europe/Berlin:{0}
DTEND;TZID=Europe/Berlin:{1}
END:VEVENT
END:VCALENDAR
"""

event_rrule_this_and_future = \
    event_rrule_this_and_future_temp.format('20140707T090000', '20140707T180000')


def test_event_rrule_this_and_future():
    dbi = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    dbi.update(event_rrule_this_and_future, href='12345.ics', etag='abcd', calendar=calname)
    assert dbi.list(calname) == [('12345.ics', 'abcd')]
    events = dbi.get_localized(BERLIN.localize(dt.datetime(2014, 4, 30, 0, 0)),
                               BERLIN.localize(dt.datetime(2014, 9, 26, 0, 0)))
    events = sorted(events, key=itemgetter(2))
    assert len(events) == 6

    assert events[0][2] == BERLIN.localize(dt.datetime(2014, 6, 30, 7, 0))
    assert events[1][2] == BERLIN.localize(dt.datetime(2014, 7, 7, 9, 0))
    assert events[2][2] == BERLIN.localize(dt.datetime(2014, 7, 14, 9, 0))
    assert events[3][2] == BERLIN.localize(dt.datetime(2014, 7, 21, 9, 0))
    assert events[4][2] == BERLIN.localize(dt.datetime(2014, 7, 28, 9, 0))
    assert events[5][2] == BERLIN.localize(dt.datetime(2014, 8, 4, 9, 0))

    assert events[0][3] == BERLIN.localize(dt.datetime(2014, 6, 30, 12, 0))
    assert events[1][3] == BERLIN.localize(dt.datetime(2014, 7, 7, 18, 0))
    assert events[2][3] == BERLIN.localize(dt.datetime(2014, 7, 14, 18, 0))
    assert events[3][3] == BERLIN.localize(dt.datetime(2014, 7, 21, 18, 0))
    assert events[4][3] == BERLIN.localize(dt.datetime(2014, 7, 28, 18, 0))
    assert events[5][3] == BERLIN.localize(dt.datetime(2014, 8, 4, 18, 0))

    assert 'SUMMARY:Arbeit\n' in events[0][0]
    for event in events[1:]:
        assert 'SUMMARY:Arbeit (lang)\n' in event[0]


event_rrule_this_and_future_multi_day_shift = \
    event_rrule_this_and_future_temp.format('20140708T090000', '20140709T150000')


def test_event_rrule_this_and_future_multi_day_shift():
    dbi = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    dbi.update(event_rrule_this_and_future_multi_day_shift,
               href='12345.ics', etag='abcd', calendar=calname)
    assert dbi.list(calname) == [('12345.ics', 'abcd')]
    events = dbi.get_localized(BERLIN.localize(dt.datetime(2014, 4, 30, 0, 0)),
                               BERLIN.localize(dt.datetime(2014, 9, 26, 0, 0)))
    events = sorted(events, key=itemgetter(2))
    assert len(events) == 6

    assert events[0][2] == BERLIN.localize(dt.datetime(2014, 6, 30, 7, 0))
    assert events[1][2] == BERLIN.localize(dt.datetime(2014, 7, 8, 9, 0))
    assert events[2][2] == BERLIN.localize(dt.datetime(2014, 7, 15, 9, 0))
    assert events[3][2] == BERLIN.localize(dt.datetime(2014, 7, 22, 9, 0))
    assert events[4][2] == BERLIN.localize(dt.datetime(2014, 7, 29, 9, 0))
    assert events[5][2] == BERLIN.localize(dt.datetime(2014, 8, 5, 9, 0))

    assert events[0][3] == BERLIN.localize(dt.datetime(2014, 6, 30, 12, 0))
    assert events[1][3] == BERLIN.localize(dt.datetime(2014, 7, 9, 15, 0))
    assert events[2][3] == BERLIN.localize(dt.datetime(2014, 7, 16, 15, 0))
    assert events[3][3] == BERLIN.localize(dt.datetime(2014, 7, 23, 15, 0))
    assert events[4][3] == BERLIN.localize(dt.datetime(2014, 7, 30, 15, 0))
    assert events[5][3] == BERLIN.localize(dt.datetime(2014, 8, 6, 15, 0))

    assert 'SUMMARY:Arbeit\n' in events[0][0]
    for event in events[1:]:
        assert 'SUMMARY:Arbeit (lang)\n' in event[0]


event_rrule_this_and_future_allday_temp = """
BEGIN:VCALENDAR
BEGIN:VEVENT
UID:event_rrule_recurrence_id_allday
SUMMARY:Arbeit
RRULE:FREQ=WEEKLY;UNTIL=20140806
DTSTART;VALUE=DATE:20140630
DTEND;VALUE=DATE:20140701
END:VEVENT
BEGIN:VEVENT
UID:event_rrule_recurrence_id_allday
SUMMARY:Arbeit (lang)
RECURRENCE-ID;RANGE=THISANDFUTURE;VALUE=DATE:20140707
DTSTART;VALUE=DATE:{}
DTEND;VALUE=DATE:{}
END:VEVENT
END:VCALENDAR
"""

event_rrule_this_and_future_allday = \
    event_rrule_this_and_future_allday_temp.format(20140708, 20140709)


def test_event_rrule_this_and_future_allday():
    dbi = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    dbi.update(event_rrule_this_and_future_allday,
               href='rrule_this_and_future_allday.ics', etag='abcd', calendar=calname)
    assert dbi.list(calname) == [('rrule_this_and_future_allday.ics', 'abcd')]
    events = list(dbi.get_floating(dt.datetime(2014, 4, 30, 0, 0), dt.datetime(2014, 9, 27, 0, 0)))
    assert len(events) == 6

    assert events[0][2] == dt.date(2014, 6, 30)
    assert events[1][2] == dt.date(2014, 7, 8)
    assert events[2][2] == dt.date(2014, 7, 15)
    assert events[3][2] == dt.date(2014, 7, 22)
    assert events[4][2] == dt.date(2014, 7, 29)
    assert events[5][2] == dt.date(2014, 8, 5)

    assert events[0][3] == dt.date(2014, 7, 1)
    assert events[1][3] == dt.date(2014, 7, 9)
    assert events[2][3] == dt.date(2014, 7, 16)
    assert events[3][3] == dt.date(2014, 7, 23)
    assert events[4][3] == dt.date(2014, 7, 30)
    assert events[5][3] == dt.date(2014, 8, 6)

    assert 'SUMMARY:Arbeit\n' in events[0][0]
    for event in events[1:]:
        assert 'SUMMARY:Arbeit (lang)\n' in event[0]


def test_event_rrule_this_and_future_allday_prior():
    event_rrule_this_and_future_allday_prior = \
        event_rrule_this_and_future_allday_temp.format(20140705, 20140706)
    dbi = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    dbi.update(event_rrule_this_and_future_allday_prior,
               href='rrule_this_and_future_allday.ics', etag='abcd', calendar=calname)
    assert dbi.list(calname) == [('rrule_this_and_future_allday.ics', 'abcd')]
    events = list(dbi.get_floating(dt.datetime(2014, 4, 30, 0, 0), dt.datetime(2014, 9, 27, 0, 0)))

    assert len(events) == 6

    assert events[0][2] == dt.date(2014, 6, 30)
    assert events[1][2] == dt.date(2014, 7, 5)
    assert events[2][2] == dt.date(2014, 7, 12)
    assert events[3][2] == dt.date(2014, 7, 19)
    assert events[4][2] == dt.date(2014, 7, 26)
    assert events[5][2] == dt.date(2014, 8, 2)

    assert events[0][3] == dt.date(2014, 7, 1)
    assert events[1][3] == dt.date(2014, 7, 6)
    assert events[2][3] == dt.date(2014, 7, 13)
    assert events[3][3] == dt.date(2014, 7, 20)
    assert events[4][3] == dt.date(2014, 7, 27)
    assert events[5][3] == dt.date(2014, 8, 3)

    assert 'SUMMARY:Arbeit\n' in events[0][0]
    for event in events[1:]:
        assert 'SUMMARY:Arbeit (lang)\n' in event[0]


event_rrule_multi_this_and_future_allday = """BEGIN:VCALENDAR
BEGIN:VEVENT
UID:event_multi_rrule_recurrence_id_allday
SUMMARY:Arbeit
RRULE:FREQ=WEEKLY;UNTIL=20140806
DTSTART;VALUE=DATE:20140630
DTEND;VALUE=DATE:20140701
END:VEVENT
BEGIN:VEVENT
UID:event_multi_rrule_recurrence_id_allday
SUMMARY:Arbeit (neu)
RECURRENCE-ID;RANGE=THISANDFUTURE;VALUE=DATE:20140721
DTSTART;VALUE=DATE:20140717
DTEND;VALUE=DATE:20140718
END:VEVENT
BEGIN:VEVENT
UID:event_multi_rrule_recurrence_id_allday
SUMMARY:Arbeit (lang)
RECURRENCE-ID;RANGE=THISANDFUTURE;VALUE=DATE:20140707
DTSTART;VALUE=DATE:20140712
DTEND;VALUE=DATE:20140714
END:VEVENT
END:VCALENDAR"""


def test_event_rrule_multi_this_and_future_allday():
    dbi = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    dbi.update(event_rrule_multi_this_and_future_allday,
               href='event_rrule_multi_this_and_future_allday.ics', etag='abcd', calendar=calname)
    assert dbi.list(calname) == [('event_rrule_multi_this_and_future_allday.ics', 'abcd')]
    events = sorted(
        dbi.get_floating(dt.datetime(2014, 4, 30, 0, 0), dt.datetime(2014, 9, 27, 0, 0)),
    )
    assert len(events) == 6

    assert events[0][2] == dt.date(2014, 6, 30)
    assert events[1][2] == dt.date(2014, 7, 12)
    assert events[2][2] == dt.date(2014, 7, 17)
    assert events[3][2] == dt.date(2014, 7, 19)
    assert events[4][2] == dt.date(2014, 7, 24)
    assert events[5][2] == dt.date(2014, 7, 31)

    assert events[0][3] == dt.date(2014, 7, 1)
    assert events[1][3] == dt.date(2014, 7, 14)
    assert events[2][3] == dt.date(2014, 7, 18)
    assert events[3][3] == dt.date(2014, 7, 21)
    assert events[4][3] == dt.date(2014, 7, 25)
    assert events[5][3] == dt.date(2014, 8, 1)

    assert 'SUMMARY:Arbeit\n' in events[0][0]
    for event in [events[1], events[3]]:
        assert 'SUMMARY:Arbeit (lang)\n' in event[0]
    for event in [events[2], events[4], events[5]]:
        assert 'SUMMARY:Arbeit (neu)\n' in event[0]


master = """BEGIN:VEVENT
UID:event_rrule_recurrence_id
SUMMARY:Arbeit
RRULE:FREQ=WEEKLY;UNTIL=20140806T060000Z
DTSTART;TZID=Europe/Berlin:20140630T070000
DTEND;TZID=Europe/Berlin:20140630T120000
END:VEVENT"""

recuid_this_future = icalendar.Event.from_ical("""BEGIN:VEVENT
UID:event_rrule_recurrence_id
SUMMARY:Arbeit
RECURRENCE-ID;RANGE=THISANDFUTURE:20140707T050000Z
DTSTART;TZID=Europe/Berlin:20140707T090000
DTEND;TZID=Europe/Berlin:20140707T140000
END:VEVENT""")

recuid_this_future_duration = icalendar.Event.from_ical("""BEGIN:VEVENT
UID:event_rrule_recurrence_id
SUMMARY:Arbeit
RECURRENCE-ID;RANGE=THISANDFUTURE:20140707T050000Z
DTSTART;TZID=Europe/Berlin:20140707T090000
DURATION:PT4H30M
END:VEVENT""")


def test_calc_shift_deltas():
    assert (dt.timedelta(hours=2), dt.timedelta(hours=5)) == \
        backend.calc_shift_deltas(recuid_this_future)
    assert (dt.timedelta(hours=2), dt.timedelta(hours=4, minutes=30)) == \
        backend.calc_shift_deltas(recuid_this_future_duration)


event_a = """BEGIN:VEVENT
UID:123
SUMMARY:event a
RRULE:FREQ=WEEKLY;UNTIL=20140806T060000Z
DTSTART;TZID=Europe/Berlin:20140630T070000
DTEND;TZID=Europe/Berlin:20140630T120000
END:VEVENT"""

event_b = """BEGIN:VEVENT
UID:123
SUMMARY:event b
RRULE:FREQ=WEEKLY;UNTIL=20140806T060000Z
DTSTART;TZID=Europe/Berlin:20140630T070000
DTEND;TZID=Europe/Berlin:20140630T120000
END:VEVENT"""


def test_two_calendars_same_uid():
    home = 'home'
    work = 'work'
    dbi = backend.SQLiteDb([home, work], ':memory:', locale=LOCALE_BERLIN)
    assert dbi.list(home) == []
    assert dbi.list(work) == []
    dbi.update(event_a, href='12345.ics', etag='abcd', calendar=home)
    assert dbi.list(home) == [('12345.ics', 'abcd')]
    assert dbi.list(work) == []
    dbi.update(event_b, href='12345.ics', etag='abcd', calendar=work)
    assert dbi.list(home) == [('12345.ics', 'abcd')]
    assert dbi.list(work) == [('12345.ics', 'abcd')]
    dbi.calendars = [home]
    events_a = list(dbi.get_localized(BERLIN.localize(dt.datetime(2014, 6, 30, 0, 0)),
                                      BERLIN.localize(dt.datetime(2014, 7, 26, 0, 0))))
    dbi.calendars = [work]
    events_b = list(dbi.get_localized(BERLIN.localize(dt.datetime(2014, 6, 30, 0, 0)),
                                      BERLIN.localize(dt.datetime(2014, 7, 26, 0, 0))))
    assert len(events_a) == 4
    assert len(events_b) == 4
    dbi.calendars = [work, home]
    events_c = list(dbi.get_localized(BERLIN.localize(dt.datetime(2014, 6, 30, 0, 0)),
                                      BERLIN.localize(dt.datetime(2014, 7, 26, 0, 0))))
    assert len(events_c) == 8
    # count events from a given calendar
    assert [event[6] for event in events_c].count(home) == 4
    assert [event[6] for event in events_c].count(work) == 4

    dbi.delete('12345.ics', calendar=home)
    dbi.calendars = [home]
    events_a = list(dbi.get_localized(BERLIN.localize(dt.datetime(2014, 6, 30, 0, 0)),
                                      BERLIN.localize(dt.datetime(2014, 7, 26, 0, 0))))
    dbi.calendars = [work]
    events_b = list(dbi.get_localized(BERLIN.localize(dt.datetime(2014, 6, 30, 0, 0)),
                                      BERLIN.localize(dt.datetime(2014, 7, 26, 0, 0))))
    assert len(events_a) == 0
    assert len(events_b) == 4
    dbi.calendars = [work, home]
    events_c = list(dbi.get_localized(BERLIN.localize(dt.datetime(2014, 6, 30, 0, 0)),
                                      BERLIN.localize(dt.datetime(2014, 7, 26, 0, 0))))
    assert [event[6] for event in events_c].count('home') == 0
    assert [event[6] for event in events_c].count('work') == 4
    assert dbi.list(home) == []
    assert dbi.list(work) == [('12345.ics', 'abcd')]


def test_update_one_should_not_affect_others():
    """test if an THISANDFUTURE param effects other events as well"""
    db = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    db.update(_get_text('event_d_15'), href='first', calendar=calname)
    events = db.get_floating(dt.datetime(2015, 4, 9, 0, 0), dt.datetime(2015, 4, 10, 0, 0))
    assert len(list(events)) == 1
    db.update(event_rrule_multi_this_and_future_allday, href='second', calendar=calname)
    events = list(db.get_floating(dt.datetime(2015, 4, 9, 0, 0), dt.datetime(2015, 4, 10, 0, 0)))
    assert len(events) == 1


def test_no_dtend():
    """test support for events with no dtend"""
    db = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    db.update(_get_text('event_dt_no_end'), href='event_dt_no_end', calendar=calname)
    events = db.get_localized(
        BERLIN.localize(dt.datetime(2016, 1, 16, 0, 0)),
        BERLIN.localize(dt.datetime(2016, 1, 17, 0, 0)),
    )
    event = list(events)[0]
    assert event[2] == BERLIN.localize(dt.datetime(2016, 1, 16, 8, 0))
    assert event[3] == BERLIN.localize(dt.datetime(2016, 1, 16, 9, 0))


event_rdate_period = """BEGIN:VEVENT
SUMMARY:RDATE period
DTSTART:19961230T020000Z
DTEND:19961230T060000Z
UID:rdate_period
RDATE;VALUE=PERIOD:19970101T180000Z/19970102T070000Z,19970109T180000Z/PT5H30M
END:VEVENT"""


supported_events = [
    event_a, event_b, event_rrule_this_and_future,
    event_rrule_this_and_future_allday,
    event_rrule_this_and_future_multi_day_shift
]


def test_check_support():
    for cal_str in supported_events:
        ical = icalendar.Calendar.from_ical(cal_str)
        [backend.check_support(event, '', '') for event in ical.walk()]

    ical = icalendar.Calendar.from_ical(event_rrule_this_and_prior)
    with pytest.raises(UpdateFailed):
        [backend.check_support(event, '', '') for event in ical.walk()]

    # icalendar 3.9.2 changed how it deals with unsupported components
    if pkg_resources.get_distribution('icalendar').parsed_version \
       <= pkg_resources.parse_version('3.9.1'):
        ical = icalendar.Calendar.from_ical(event_rdate_period)
        with pytest.raises(UpdateFailed):
            [backend.check_support(event, '', '') for event in ical.walk()]


def test_check_support_rdate_no_values():
    """check if `check_support` doesn't choke on events with an RDATE property
    without a VALUE parameter"""
    ical = icalendar.Calendar.from_ical(_get_text('event_rdate_no_value'))
    [backend.check_support(event, '', '') for event in ical.walk()]


card = """BEGIN:VCARD
VERSION:3.0
FN:Unix
BDAY:19710311
END:VCARD
"""

card_29thfeb = """BEGIN:VCARD
VERSION:3.0
FN:leapyear
BDAY:20000229
END:VCARD
"""

card_no_year = """BEGIN:VCARD
VERSION:3.0
FN:Unix
BDAY:--0311
END:VCARD
"""

card_does_not_parse = """BEGIN:VCARD
VERSION:3.0
FN:Unix
BDAY:x
END:VCARD
"""

card_no_fn = """BEGIN:VCARD
VERSION:3.0
N:Ritchie;Dennis;MacAlistair;;
BDAY:19410909
END:VCARD
"""

card_two_birthdays = """BEGIN:VCARD
VERSION:3.0
N:Ritchie;Dennis;MacAlistair;;
BDAY:19410909
BDAY:--0311
END:VCARD
"""

card_anniversary = """BEGIN:VCARD
VERSION:3.0
FN:Unix
X-ANNIVERSARY:19710311
END:VCARD
"""

card_abdate = """BEGIN:VCARD
VERSION:3.0
FN:Unix
ITEM1.X-ABDATE:19710311
ITEM1.X-ABLabel:spouse's birthday
END:VCARD
"""

card_abdate_nolabel = """BEGIN:VCARD
VERSION:3.0
FN:Unix
ITEM1.X-ABDATE:19710311
END:VCARD
"""

card_v3 = """BEGIN:VCARD
VERSION:3.0
FN:Unix
BDAY:1971-03-11
END:VCARD
"""

day = dt.date(1971, 3, 11)
start = dt.datetime.combine(day, dt.time.min)
end = dt.datetime.combine(day, dt.time.max)


def test_birthdays():
    db = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    assert list(db.get_floating(start, end)) == []
    db.update_vcf_dates(card, 'unix.vcf', calendar=calname)
    events = list(db.get_floating(start, end))
    assert len(events) == 1
    assert 'SUMMARY:Unix\'s birthday' in events[0][0]

    events = list(
        db.get_floating(
            dt.datetime(2016, 3, 11, 0, 0),
            dt.datetime(2016, 3, 11, 23, 59, 59, 999)))
    assert 'SUMMARY:Unix\'s birthday' in events[0][0]


def test_birthdays_update():
    """test if we can update a birthday"""
    db = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    db.update_vcf_dates(card, 'unix.vcf', calendar=calname)
    db.update_vcf_dates(card, 'unix.vcf', calendar=calname)


def test_birthdays_no_fn():
    db = backend.SQLiteDb(['home'], ':memory:', locale=LOCALE_BERLIN)
    assert list(db.get_floating(dt.datetime(1941, 9, 9, 0, 0),
                                dt.datetime(1941, 9, 9, 23, 59, 59, 9999))) == []
    db.update_vcf_dates(card_no_fn, 'unix.vcf', calendar=calname)
    events = list(db.get_floating(dt.datetime(1941, 9, 9, 0, 0),
                                  dt.datetime(1941, 9, 9, 23, 59, 59, 9999)))
    assert len(events) == 1
    assert 'SUMMARY:Dennis MacAlistair Ritchie\'s birthday' in events[0][0]


def test_birthday_does_not_parse():
    db = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    assert list(db.get_floating(start, end)) == []
    db.update_vcf_dates(card_does_not_parse, 'unix.vcf', calendar=calname)
    events = list(db.get_floating(start, end))
    assert len(events) == 0


def test_vcard_two_birthdays():
    db = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    assert list(db.get_floating(start, end)) == []
    db.update_vcf_dates(card_two_birthdays, 'unix.vcf', calendar=calname)
    events = list(db.get_floating(start, end))
    assert len(events) == 0


def test_anniversary():
    db = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    assert list(db.get_floating(start, end)) == []
    db.update_vcf_dates(card_anniversary, 'unix.vcf', calendar=calname)
    events = list(db.get_floating(start, end))
    assert len(events) == 1
    assert 'SUMMARY:Unix\'s anniversary' in events[0][0]

    events = list(
        db.get_floating(
            dt.datetime(2016, 3, 11, 0, 0),
            dt.datetime(2016, 3, 11, 23, 59, 59, 999)))
    assert 'SUMMARY:Unix\'s anniversary' in events[0][0]


def test_abdate():
    db = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    assert list(db.get_floating(start, end)) == []
    db.update_vcf_dates(card_abdate, 'unix.vcf', calendar=calname)
    events = list(db.get_floating(start, end))
    assert len(events) == 1
    assert 'SUMMARY:Unix\'s spouse\'s birthday' in events[0][0]

    events = list(
        db.get_floating(
            dt.datetime(2016, 3, 11, 0, 0),
            dt.datetime(2016, 3, 11, 23, 59, 59, 999)))
    assert 'SUMMARY:Unix\'s spouse\'s birthday' in events[0][0]


def test_abdate_nolabel():
    db = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    assert list(db.get_floating(start, end)) == []
    db.update_vcf_dates(card_abdate_nolabel, 'unix.vcf', calendar=calname)
    events = list(db.get_floating(start, end))
    assert len(events) == 1
    assert 'SUMMARY:Unix\'s custom event from vcard' in events[0][0]

    events = list(
        db.get_floating(
            dt.datetime(2016, 3, 11, 0, 0),
            dt.datetime(2016, 3, 11, 23, 59, 59, 999)))
    assert 'SUMMARY:Unix\'s custom event from vcard' in events[0][0]


def test_birthday_v3():
    db = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
    assert list(db.get_floating(start, end)) == []
    db.update_vcf_dates(card_v3, 'unix.vcf', calendar=calname)
    events = list(db.get_floating(start, end))
    assert len(events) == 1
    assert 'SUMMARY:Unix\'s birthday' in events[0][0]

    events = list(
        db.get_floating(
            dt.datetime(2016, 3, 11, 0, 0),
            dt.datetime(2016, 3, 11, 23, 59, 59, 999)))
    assert 'SUMMARY:Unix\'s birthday' in events[0][0]



import datetime as dt

import icalendar
import pytz

from khal import icalendar as icalendar_helpers
from khal import utils

from .utils import _get_text, _get_vevent_file

# FIXME this file is in urgent need of a clean up

BERLIN = pytz.timezone('Europe/Berlin')
BOGOTA = pytz.timezone('America/Bogota')

# datetime
event_dt = """BEGIN:VCALENDAR
CALSCALE:GREGORIAN
VERSION:2.0
BEGIN:VEVENT
SUMMARY:Datetime Event
DTSTART;TZID=Europe/Berlin;VALUE=DATE-TIME:20130301T140000
DTEND;TZID=Europe/Berlin;VALUE=DATE-TIME:20130301T160000
RRULE:FREQ=MONTHLY;INTERVAL=2;COUNT=6
UID:datetime123
END:VEVENT
END:VCALENDAR"""

event_dt_norr = """BEGIN:VCALENDAR
CALSCALE:GREGORIAN
VERSION:2.0
BEGIN:VEVENT
SUMMARY:Datetime Event
DTSTART;TZID=Europe/Berlin;VALUE=DATE-TIME:20130301T140000
DTEND;TZID=Europe/Berlin;VALUE=DATE-TIME:20130301T160000
UID:datetime123
END:VEVENT
END:VCALENDAR"""

# datetime zulu (in utc time)
event_dttz = """BEGIN:VCALENDAR
CALSCALE:GREGORIAN
VERSION:2.0
BEGIN:VEVENT
SUMMARY:Datetime Zulu Event
DTSTART;VALUE=DATE-TIME:20130301T140000Z
DTEND;VALUE=DATE-TIME:20130301T160000Z
RRULE:FREQ=MONTHLY;INTERVAL=2;COUNT=6
UID:datetimezulu123
END:VEVENT
END:VCALENDAR"""

event_dttz_norr = """BEGIN:VCALENDAR
CALSCALE:GREGORIAN
VERSION:2.0
BEGIN:VEVENT
SUMMARY:Datetime Zulu Event
DTSTART;VALUE=DATE-TIME:20130301T140000Z
DTEND;VALUE=DATE-TIME:20130301T160000Z
UID:datetimezulu123
END:VEVENT
END:VCALENDAR"""

# datetime floating (no time zone information)
event_dtf = """BEGIN:VCALENDAR
CALSCALE:GREGORIAN
VERSION:2.0
BEGIN:VEVENT
SUMMARY:Datetime floating Event
DTSTART;VALUE=DATE-TIME:20130301T140000
DTEND;VALUE=DATE-TIME:20130301T160000
RRULE:FREQ=MONTHLY;INTERVAL=2;COUNT=6
UID:datetimefloating123
END:VEVENT
END:VCALENDAR"""

event_dtf_norr = """BEGIN:VCALENDAR
CALSCALE:GREGORIAN
VERSION:2.0
BEGIN:VEVENT
SUMMARY:Datetime floating Event
DTSTART;VALUE=DATE-TIME:20130301T140000
DTEND;VALUE=DATE-TIME:20130301T160000
UID:datetimefloating123
END:VEVENT
END:VCALENDAR"""

# datetime broken (as in we don't understand the timezone information)
event_dtb = """BEGIN:VCALENDAR
CALSCALE:GREGORIAN
VERSION:2.0
BEGIN:VTIMEZONE
TZID:/freeassociation.sourceforge.net/Tzfile/Europe/Berlin
X-LIC-LOCATION:Europe/Berlin
BEGIN:STANDARD
TZNAME:CET
DTSTART:19701027T030000
RRULE:FREQ=YEARLY;BYDAY=-1SU;BYMONTH=10
TZOFFSETFROM:+0200
TZOFFSETTO:+0100
END:STANDARD
BEGIN:DAYLIGHT
TZNAME:CEST
DTSTART:19700331T020000
RRULE:FREQ=YEARLY;BYDAY=-1SU;BYMONTH=3
TZOFFSETFROM:+0100
TZOFFSETTO:+0200
END:DAYLIGHT
END:VTIMEZONE
BEGIN:VEVENT
UID:broken123
DTSTART;TZID=/freeassociation.sourceforge.net/Tzfile/Europe/Berlin:20130301T140000
DTEND;TZID=/freeassociation.sourceforge.net/Tzfile/Europe/Berlin:20130301T160000
RRULE:FREQ=MONTHLY;INTERVAL=2;COUNT=6
TRANSP:OPAQUE
SEQUENCE:2
SUMMARY:Broken Event
END:VEVENT
END:VCALENDAR
"""

event_dtb_norr = """BEGIN:VCALENDAR
CALSCALE:GREGORIAN
VERSION:2.0
BEGIN:VTIMEZONE
TZID:/freeassociation.sourceforge.net/Tzfile/Europe/Berlin
X-LIC-LOCATION:Europe/Berlin
BEGIN:STANDARD
TZNAME:CET
DTSTART:19701027T030000
RRULE:FREQ=YEARLY;BYDAY=-1SU;BYMONTH=10
TZOFFSETFROM:+0200
TZOFFSETTO:+0100
END:STANDARD
BEGIN:DAYLIGHT
TZNAME:CEST
DTSTART:19700331T020000
RRULE:FREQ=YEARLY;BYDAY=-1SU;BYMONTH=3
TZOFFSETFROM:+0100
TZOFFSETTO:+0200
END:DAYLIGHT
END:VTIMEZONE
BEGIN:VEVENT
UID:broken123
DTSTART;TZID=/freeassociation.sourceforge.net/Tzfile/Europe/Berlin:20130301T140000
DTEND;TZID=/freeassociation.sourceforge.net/Tzfile/Europe/Berlin:20130301T160000
TRANSP:OPAQUE
SEQUENCE:2
SUMMARY:Broken Event
END:VEVENT
END:VCALENDAR
"""

# all day (date) event
event_d = """BEGIN:VCALENDAR
CALSCALE:GREGORIAN
VERSION:2.0
BEGIN:VEVENT
UID:date123
DTSTART;VALUE=DATE:20130301
DTEND;VALUE=DATE:20130302
RRULE:FREQ=MONTHLY;INTERVAL=2;COUNT=6
SUMMARY:Event
END:VEVENT
END:VCALENDAR
"""

# all day (date) event with timezone information
event_dtz = """BEGIN:VCALENDAR
CALSCALE:GREGORIAN
VERSION:2.0
BEGIN:VEVENT
UID:datetz123
DTSTART;TZID=Berlin/Europe;VALUE=DATE:20130301
DTEND;TZID=Berlin/Europe;VALUE=DATE:20130302
RRULE:FREQ=MONTHLY;INTERVAL=2;COUNT=6
SUMMARY:Event
END:VEVENT
END:VCALENDAR
"""

event_dtzb = """BEGIN:VCALENDAR
CALSCALE:GREGORIAN
VERSION:2.0
BEGIN:VTIMEZONE
TZID:Pacific Time (US & Canada), Tijuana
BEGIN:STANDARD
TZNAME:PST
DTSTART:20071104T020000
TZOFFSETTO:-0800
TZOFFSETFROM:-0700
RRULE:FREQ=YEARLY;BYMONTH=11;BYDAY=1SU
END:STANDARD
BEGIN:DAYLIGHT
TZNAME:PDT
DTSTART:20070311T020000
TZOFFSETTO:-0700
TZOFFSETFROM:-0800
RRULE:FREQ=YEARLY;BYMONTH=3;BYDAY=2SU
END:DAYLIGHT
END:VTIMEZONE
BEGIN:VEVENT
DTSTART;VALUE=DATE;TZID="Pacific Time (US & Canada), Tijuana":20130301
DTEND;VALUE=DATE;TZID="Pacific Time (US & Canada), Tijuana":20130302
RRULE:FREQ=MONTHLY;INTERVAL=2;COUNT=6
SUMMARY:Event
UID:eventdtzb123
END:VEVENT
END:VCALENDAR
"""

event_d_norr = """BEGIN:VCALENDAR
CALSCALE:GREGORIAN
VERSION:2.0
BEGIN:VEVENT
UID:date123
DTSTART;VALUE=DATE:20130301
DTEND;VALUE=DATE:20130302
SUMMARY:Event
END:VEVENT
END:VCALENDAR
"""

berlin = pytz.timezone('Europe/Berlin')
new_york = pytz.timezone('America/New_York')


def _get_vevent(event):
    ical = icalendar.Event.from_ical(event)
    for component in ical.walk():
        if component.name == 'VEVENT':
            return component


class TestExpand:
    dtstartend_berlin = [
        (berlin.localize(dt.datetime(2013, 3, 1, 14, 0, )),
         berlin.localize(dt.datetime(2013, 3, 1, 16, 0, ))),
        (berlin.localize(dt.datetime(2013, 5, 1, 14, 0, )),
         berlin.localize(dt.datetime(2013, 5, 1, 16, 0, ))),
        (berlin.localize(dt.datetime(2013, 7, 1, 14, 0, )),
         berlin.localize(dt.datetime(2013, 7, 1, 16, 0, ))),
        (berlin.localize(dt.datetime(2013, 9, 1, 14, 0, )),
         berlin.localize(dt.datetime(2013, 9, 1, 16, 0, ))),
        (berlin.localize(dt.datetime(2013, 11, 1, 14, 0,)),
         berlin.localize(dt.datetime(2013, 11, 1, 16, 0,))),
        (berlin.localize(dt.datetime(2014, 1, 1, 14, 0, )),
         berlin.localize(dt.datetime(2014, 1, 1, 16, 0, )))
    ]

    dtstartend_utc = [
        (dt.datetime(2013, 3, 1, 14, 0, tzinfo=pytz.utc),
         dt.datetime(2013, 3, 1, 16, 0, tzinfo=pytz.utc)),
        (dt.datetime(2013, 5, 1, 14, 0, tzinfo=pytz.utc),
         dt.datetime(2013, 5, 1, 16, 0, tzinfo=pytz.utc)),
        (dt.datetime(2013, 7, 1, 14, 0, tzinfo=pytz.utc),
         dt.datetime(2013, 7, 1, 16, 0, tzinfo=pytz.utc)),
        (dt.datetime(2013, 9, 1, 14, 0, tzinfo=pytz.utc),
         dt.datetime(2013, 9, 1, 16, 0, tzinfo=pytz.utc)),
        (dt.datetime(2013, 11, 1, 14, 0, tzinfo=pytz.utc),
         dt.datetime(2013, 11, 1, 16, 0, tzinfo=pytz.utc)),
        (dt.datetime(2014, 1, 1, 14, 0, tzinfo=pytz.utc),
         dt.datetime(2014, 1, 1, 16, 0, tzinfo=pytz.utc))
    ]

    dtstartend_float = [
        (dt.datetime(2013, 3, 1, 14, 0),
         dt.datetime(2013, 3, 1, 16, 0)),
        (dt.datetime(2013, 5, 1, 14, 0),
         dt.datetime(2013, 5, 1, 16, 0)),
        (dt.datetime(2013, 7, 1, 14, 0),
         dt.datetime(2013, 7, 1, 16, 0)),
        (dt.datetime(2013, 9, 1, 14, 0),
         dt.datetime(2013, 9, 1, 16, 0)),
        (dt.datetime(2013, 11, 1, 14, 0),
         dt.datetime(2013, 11, 1, 16, 0)),
        (dt.datetime(2014, 1, 1, 14, 0),
         dt.datetime(2014, 1, 1, 16, 0))
    ]
    dstartend = [
        (dt.date(2013, 3, 1,),
         dt.date(2013, 3, 2,)),
        (dt.date(2013, 5, 1,),
         dt.date(2013, 5, 2,)),
        (dt.date(2013, 7, 1,),
         dt.date(2013, 7, 2,)),
        (dt.date(2013, 9, 1,),
         dt.date(2013, 9, 2,)),
        (dt.date(2013, 11, 1),
         dt.date(2013, 11, 2)),
        (dt.date(2014, 1, 1,),
         dt.date(2014, 1, 2,))
    ]
    offset_berlin = [
        dt.timedelta(0, 3600),
        dt.timedelta(0, 7200),
        dt.timedelta(0, 7200),
        dt.timedelta(0, 7200),
        dt.timedelta(0, 3600),
        dt.timedelta(0, 3600)
    ]

    offset_utc = [
        dt.timedelta(0, 0),
        dt.timedelta(0, 0),
        dt.timedelta(0, 0),
        dt.timedelta(0, 0),
        dt.timedelta(0, 0),
        dt.timedelta(0, 0),
    ]

    offset_none = [None, None, None, None, None, None]

    def test_expand_dt(self):
        vevent = _get_vevent(event_dt)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert dtstart == self.dtstartend_berlin
        assert [start.utcoffset()
                for start, _ in dtstart] == self.offset_berlin
        assert [end.utcoffset() for _, end in dtstart] == self.offset_berlin

    def test_expand_dtb(self):
        vevent = _get_vevent(event_dtb)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert dtstart == self.dtstartend_berlin
        assert [start.utcoffset()
                for start, _ in dtstart] == self.offset_berlin
        assert [end.utcoffset() for _, end in dtstart] == self.offset_berlin

    def test_expand_dttz(self):
        vevent = _get_vevent(event_dttz)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert dtstart == self.dtstartend_utc
        assert [start.utcoffset() for start, _ in dtstart] == self.offset_utc
        assert [end.utcoffset() for _, end in dtstart] == self.offset_utc

    def test_expand_dtf(self):
        vevent = _get_vevent(event_dtf)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert dtstart == self.dtstartend_float
        assert [start.utcoffset() for start, _ in dtstart] == self.offset_none
        assert [end.utcoffset() for _, end in dtstart] == self.offset_none

    def test_expand_d(self):
        vevent = _get_vevent(event_d)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert dtstart == self.dstartend

    def test_expand_dtz(self):
        vevent = _get_vevent(event_dtz)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert dtstart == self.dstartend

    def test_expand_dtzb(self):
        vevent = _get_vevent(event_dtzb)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert dtstart == self.dstartend

    def test_expand_invalid_exdate(self):
        """testing if we can expand an event with EXDATEs that do not much
        its RRULE"""
        vevent = _get_vevent_file('event_invalid_exdate')
        dtstartl = icalendar_helpers.expand(vevent, berlin)
        # TODO test for logging message
        assert dtstartl == [
            (new_york.localize(dt.datetime(2011, 11, 12, 15, 50)),
             new_york.localize(dt.datetime(2011, 11, 12, 17, 0))),
            (new_york.localize(dt.datetime(2011, 11, 19, 15, 50)),
             new_york.localize(dt.datetime(2011, 11, 19, 17, 0))),
            (new_york.localize(dt.datetime(2011, 12, 3, 15, 50)),
             new_york.localize(dt.datetime(2011, 12, 3, 17, 0))),
        ]


class TestExpandNoRR:
    dtstartend_berlin = [
        (berlin.localize(dt.datetime(2013, 3, 1, 14, 0)),
         berlin.localize(dt.datetime(2013, 3, 1, 16, 0))),
    ]

    dtstartend_utc = [
        (dt.datetime(2013, 3, 1, 14, 0, tzinfo=pytz.utc),
         dt.datetime(2013, 3, 1, 16, 0, tzinfo=pytz.utc)),
    ]

    dtstartend_float = [
        (dt.datetime(2013, 3, 1, 14, 0),
         dt.datetime(2013, 3, 1, 16, 0)),
    ]
    offset_berlin = [
        dt.timedelta(0, 3600),
    ]

    offset_utc = [
        dt.timedelta(0, 0),
    ]

    offset_none = [None]

    def test_expand_dt(self):
        vevent = _get_vevent(event_dt_norr)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert dtstart == self.dtstartend_berlin
        assert [start.utcoffset()
                for start, _ in dtstart] == self.offset_berlin
        assert [end.utcoffset() for _, end in dtstart] == self.offset_berlin

    def test_expand_dtb(self):
        vevent = _get_vevent(event_dtb_norr)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert dtstart == self.dtstartend_berlin
        assert [start.utcoffset()
                for start, _ in dtstart] == self.offset_berlin
        assert [end.utcoffset() for _, end in dtstart] == self.offset_berlin

    def test_expand_dttz(self):
        vevent = _get_vevent(event_dttz_norr)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert dtstart == self.dtstartend_utc
        assert [start.utcoffset() for start, _ in dtstart] == self.offset_utc
        assert [end.utcoffset() for _, end in dtstart] == self.offset_utc

    def test_expand_dtf(self):
        vevent = _get_vevent(event_dtf_norr)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert dtstart == self.dtstartend_float
        assert [start.utcoffset() for start, _ in dtstart] == self.offset_none
        assert [end.utcoffset() for _, end in dtstart] == self.offset_none

    def test_expand_d(self):
        vevent = _get_vevent(event_d_norr)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert dtstart == [
            (dt.date(2013, 3, 1,),
             dt.date(2013, 3, 2,)),
        ]

    def test_expand_dtr_exdatez(self):
        """a recurring event with an EXDATE in Zulu time while DTSTART is
        localized"""
        vevent = _get_vevent_file('event_dtr_exdatez')
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert len(dtstart) == 3

    def test_expand_rrule_exdate_z(self):
        """event with not understood timezone for dtstart and zulu time form
        exdate
        """
        vevent = _get_vevent_file('event_dtr_no_tz_exdatez')
        vevent = icalendar_helpers.sanitize(vevent, berlin, '', '')
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert len(dtstart) == 5
        dtstarts = [start for start, end in dtstart]
        assert dtstarts == [
            berlin.localize(dt.datetime(2012, 4, 3, 10, 0)),
            berlin.localize(dt.datetime(2012, 5, 3, 10, 0)),
            berlin.localize(dt.datetime(2012, 7, 3, 10, 0)),
            berlin.localize(dt.datetime(2012, 8, 3, 10, 0)),
            berlin.localize(dt.datetime(2012, 9, 3, 10, 0)),
        ]

    def test_expand_rrule_notz_until_z(self):
        """event with not understood timezone for dtstart and zulu time form
        exdate
        """
        vevent = _get_vevent_file('event_dtr_notz_untilz')
        vevent = icalendar_helpers.sanitize(vevent, new_york, '', '')
        dtstart = icalendar_helpers.expand(vevent, new_york)
        assert len(dtstart) == 7
        dtstarts = [start for start, end in dtstart]
        assert dtstarts == [
            new_york.localize(dt.datetime(2012, 7, 26, 13, 0)),
            new_york.localize(dt.datetime(2012, 8, 9, 13, 0)),
            new_york.localize(dt.datetime(2012, 8, 23, 13, 0)),
            new_york.localize(dt.datetime(2012, 9, 6, 13, 0)),
            new_york.localize(dt.datetime(2012, 9, 20, 13, 0)),
            new_york.localize(dt.datetime(2012, 10, 4, 13, 0)),
            new_york.localize(dt.datetime(2012, 10, 18, 13, 0)),
        ]


vevent_until_notz = """BEGIN:VEVENT
SUMMARY:until 20. Februar
DTSTART;TZID=Europe/Berlin:20140203T070000
DTEND;TZID=Europe/Berlin:20140203T090000
UID:until_notz
RRULE:FREQ=DAILY;UNTIL=20140220T060000Z;WKST=SU
END:VEVENT
"""

vevent_count = """BEGIN:VEVENT
SUMMARY:until 20. Februar
DTSTART:20140203T070000
DTEND:20140203T090000
UID:until_notz
RRULE:FREQ=DAILY;UNTIL=20140220T070000;WKST=SU
END:VEVENT
"""

event_until_d_notz = """BEGIN:VCALENDAR
VERSION:2.0
BEGIN:VEVENT
UID:d470ef6d08
DTSTART;VALUE=DATE:20140110
DURATION:P1D
RRULE:FREQ=WEEKLY;UNTIL=20140215;INTERVAL=1;BYDAY=FR
SUMMARY:Fri
END:VEVENT
END:VCALENDAR
"""

event_exdate_dt = """BEGIN:VCALENDAR
VERSION:2.0
BEGIN:VEVENT
UID:event_exdate_dt123
DTSTAMP:20140627T162546Z
DTSTART;TZID=Europe/Berlin:20140702T190000
DTEND;TZID=Europe/Berlin:20140702T193000
SUMMARY:Test event
RRULE:FREQ=DAILY;COUNT=10
EXDATE:20140703T190000
END:VEVENT
END:VCALENDAR
"""

event_exdates_dt = """BEGIN:VCALENDAR
VERSION:2.0
BEGIN:VEVENT
UID:event_exdates_dt123
DTSTAMP:20140627T162546Z
DTSTART;TZID=Europe/Berlin:20140702T190000
DTEND;TZID=Europe/Berlin:20140702T193000
SUMMARY:Test event
RRULE:FREQ=DAILY;COUNT=10
EXDATE:20140703T190000
EXDATE:20140705T190000
END:VEVENT
END:VCALENDAR
"""

event_exdatesl_dt = """BEGIN:VCALENDAR
VERSION:2.0
BEGIN:VEVENT
UID:event_exdatesl_dt123
DTSTAMP:20140627T162546Z
DTSTART;TZID=Europe/Berlin:20140702T190000
DTEND;TZID=Europe/Berlin:20140702T193000
SUMMARY:Test event
RRULE:FREQ=DAILY;COUNT=10
EXDATE:20140703T190000
EXDATE:20140705T190000,20140707T190000
END:VEVENT
END:VCALENDAR
"""

latest_bug = """BEGIN:VCALENDAR
VERSION:2.0
BEGIN:VEVENT
SUMMARY:Reformationstag
RRULE:FREQ=YEARLY;BYMONTHDAY=31;BYMONTH=10
DTSTART;VALUE=DATE:20091031
DTEND;VALUE=DATE:20091101
END:VEVENT
END:VCALENDAR
"""

recurrence_id_with_timezone = """BEGIN:VEVENT
SUMMARY:PyCologne
DTSTART;TZID=/freeassociation.sourceforge.net/Tzfile/Europe/Berlin:20131113T190000
DTEND;TZID=/freeassociation.sourceforge.net/Tzfile/Europe/Berlin:20131113T210000
DTSTAMP:20130610T160635Z
UID:another_problem
RECURRENCE-ID;TZID=/freeassociation.sourceforge.net/Tzfile/Europe/Berlin:20131113T190000
RRULE:FREQ=MONTHLY;BYDAY=2WE;WKST=SU
TRANSP:OPAQUE
END:VEVENT
"""


class TestSpecial:
    """collection of strange test cases that don't fit anywhere else really"""

    def test_count(self):
        vevent = _get_vevent(vevent_count)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        starts = [start for start, _ in dtstart]
        assert len(starts) == 18
        assert dtstart[0][0] == dt.datetime(2014, 2, 3, 7, 0)
        assert dtstart[-1][0] == dt.datetime(2014, 2, 20, 7, 0)

    def test_until_notz(self):
        vevent = _get_vevent(vevent_until_notz)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        starts = [start for start, _ in dtstart]
        assert len(starts) == 18
        assert dtstart[0][0] == berlin.localize(
            dt.datetime(2014, 2, 3, 7, 0))
        assert dtstart[-1][0] == berlin.localize(
            dt.datetime(2014, 2, 20, 7, 0))

    def test_until_d_notz(self):
        vevent = _get_vevent(event_until_d_notz)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        starts = [start for start, _ in dtstart]
        assert len(starts) == 6
        assert dtstart[0][0] == dt.date(2014, 1, 10)
        assert dtstart[-1][0] == dt.date(2014, 2, 14)

    def test_latest_bug(self):
        vevent = _get_vevent(latest_bug)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert dtstart[0][0] == dt.date(2009, 10, 31)
        assert dtstart[-1][0] == dt.date(2037, 10, 31)

    def test_recurrence_id_with_timezone(self):
        vevent = _get_vevent(recurrence_id_with_timezone)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert len(dtstart) == 1
        assert dtstart[0][0] == berlin.localize(
            dt.datetime(2013, 11, 13, 19, 0))

    def test_event_exdate_dt(self):
        """recurring event, one date excluded via EXCLUDE"""
        vevent = _get_vevent(event_exdate_dt)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert len(dtstart) == 9
        assert dtstart[0][0] == berlin.localize(
            dt.datetime(2014, 7, 2, 19, 0))
        assert dtstart[-1][0] == berlin.localize(
            dt.datetime(2014, 7, 11, 19, 0))

    def test_event_exdates_dt(self):
        """recurring event, two dates excluded via EXCLUDE"""
        vevent = _get_vevent(event_exdates_dt)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert len(dtstart) == 8
        assert dtstart[0][0] == berlin.localize(
            dt.datetime(2014, 7, 2, 19, 0))
        assert dtstart[-1][0] == berlin.localize(
            dt.datetime(2014, 7, 11, 19, 0))

    def test_event_exdatesl_dt(self):
        """recurring event, three dates exclude via two EXCLUDEs"""
        vevent = _get_vevent(event_exdatesl_dt)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert len(dtstart) == 7
        assert dtstart[0][0] == berlin.localize(
            dt.datetime(2014, 7, 2, 19, 0))
        assert dtstart[-1][0] == berlin.localize(
            dt.datetime(2014, 7, 11, 19, 0))

    def test_event_exdates_remove(self):
        """check if we can remove one more instance"""
        vevent = _get_vevent(event_exdatesl_dt)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert len(dtstart) == 7

        exdate1 = pytz.UTC.localize(dt.datetime(2014, 7, 11, 17, 0))
        icalendar_helpers.delete_instance(vevent, exdate1)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert len(dtstart) == 6

        exdate2 = berlin.localize(dt.datetime(2014, 7, 9, 19, 0))
        icalendar_helpers.delete_instance(vevent, exdate2)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert len(dtstart) == 5

    def test_event_dt_rrule_invalid_until(self):
        """DTSTART and RRULE:UNTIL should be of the same type, but might not
        be"""
        vevent = _get_vevent(_get_text('event_dt_rrule_invalid_until'))
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert dtstart == [(dt.date(2007, 12, 1), dt.date(2007, 12, 2)),
                           (dt.date(2008, 1, 1), dt.date(2008, 1, 2)),
                           (dt.date(2008, 2, 1), dt.date(2008, 2, 2))]

    def test_event_dt_rrule_invalid_until2(self):
        """same as above, but now dtstart is of type date and until is datetime
        """
        vevent = _get_vevent(_get_text('event_dt_rrule_invalid_until2'))
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert len(dtstart) == 35
        assert dtstart[0] == (berlin.localize(dt.datetime(2014, 4, 9, 9, 30)),
                              berlin.localize(dt.datetime(2014, 4, 9, 10, 30)))
        assert dtstart[-1] == (berlin.localize(dt.datetime(2014, 12, 3, 9, 30)),
                               berlin.localize(dt.datetime(2014, 12, 3, 10, 30)))

    def test_event_dt_rrule_until_before_start(self):
        """test handling if an RRULE's UNTIL is before the event's DTSTART"""
        vevent = _get_vevent(_get_text('event_dt_rrule_until_before_start'))
        dtstart = icalendar_helpers.expand(vevent, berlin)
        # TODO test for logging message
        assert dtstart is None

    def test_event_invalid_rrule(self):
        """test handling if an event with RRULE will never occur"""
        vevent = _get_vevent(_get_text('event_rrule_no_occurence'))
        dtstart = icalendar_helpers.expand(vevent, berlin)
        # TODO test for logging message
        assert dtstart is None


simple_rdate = """BEGIN:VEVENT
SUMMARY:Simple Rdate
DTSTART;TZID=Europe/Berlin:20131113T190000
DTEND;TZID=Europe/Berlin:20131113T210000
UID:simple_rdate
RDATE:20131213T190000
RDATE:20140113T190000,20140213T190000
END:VEVENT
"""

rrule_and_rdate = """BEGIN:VEVENT
SUMMARY:Datetime Event
DTSTART;TZID=Europe/Berlin;VALUE=DATE-TIME:20130301T140000
DTEND;TZID=Europe/Berlin;VALUE=DATE-TIME:20130301T160000
RRULE:FREQ=MONTHLY;INTERVAL=2;COUNT=6
RDATE:20131213T190000
UID:datetime123
END:VEVENT"""


class TestRDate:
    """Testing expanding of recurrence rules"""
    def test_simple_rdate(self):
        vevent = _get_vevent(simple_rdate)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert len(dtstart) == 4

    def test_rrule_and_rdate(self):
        vevent = _get_vevent(rrule_and_rdate)
        dtstart = icalendar_helpers.expand(vevent, berlin)
        assert len(dtstart) == 7

    def test_rrule_past(self):
        vevent = _get_vevent_file('event_r_past')
        assert vevent is not None
        dtstarts = icalendar_helpers.expand(vevent, berlin)
        assert len(dtstarts) == 73
        assert dtstarts[0][0] == dt.date(1965, 4, 23)
        assert dtstarts[-1][0] == dt.date(2037, 4, 23)

    def test_rdate_date(self):
        vevent = _get_vevent_file('event_d_rdate')
        dtstarts = icalendar_helpers.expand(vevent, berlin)
        assert len(dtstarts) == 4
        assert dtstarts == [(dt.date(2015, 8, 12), dt.date(2015, 8, 13)),
                            (dt.date(2015, 8, 13), dt.date(2015, 8, 14)),
                            (dt.date(2015, 8, 14), dt.date(2015, 8, 15)),
                            (dt.date(2015, 8, 15), dt.date(2015, 8, 16))]


noend_date = """
BEGIN:VCALENDAR
BEGIN:VEVENT
UID:noend123
DTSTART;VALUE=DATE:20140829
SUMMARY:No DTEND
END:VEVENT
END:VCALENDAR
"""

noend_datetime = """
BEGIN:VCALENDAR
BEGIN:VEVENT
UID:noend123
DTSTART;TZID=Europe/Berlin;VALUE=DATE-TIME:20140829T080000
SUMMARY:No DTEND
END:VEVENT
END:VCALENDAR
"""

instant = """
BEGIN:VCALENDAR
BEGIN:VEVENT
UID:instant123
DTSTART;TZID=Europe/Berlin;VALUE=DATE-TIME:20170113T010000
DTEND;TZID=Europe/Berlin;VALUE=DATE-TIME:20170113T010000
SUMMARY:Really fast event
END:VEVENT
END:VCALENDAR
"""


class TestSanitize:

    def test_noend_date(self):
        vevent = _get_vevent(noend_date)
        vevent = icalendar_helpers.sanitize(vevent, berlin, '', '')
        assert vevent['DTSTART'].dt == dt.date(2014, 8, 29)
        assert vevent['DTEND'].dt == dt.date(2014, 8, 30)

    def test_noend_datetime(self):
        vevent = _get_vevent(noend_datetime)
        vevent = icalendar_helpers.sanitize(vevent, berlin, '', '')
        assert vevent['DTSTART'].dt == BERLIN.localize(dt.datetime(2014, 8, 29, 8))
        assert vevent['DTEND'].dt == BERLIN.localize(dt.datetime(2014, 8, 29, 9))

    def test_duration(self):
        vevent = _get_vevent_file('event_dtr_exdatez')
        vevent = icalendar_helpers.sanitize(vevent, berlin, '', '')

    def test_instant(self):
        vevent = _get_vevent(instant)
        assert vevent['DTEND'].dt - vevent['DTSTART'].dt == dt.timedelta()
        vevent = icalendar_helpers.sanitize(vevent, berlin, '', '')
        assert vevent['DTEND'].dt - vevent['DTSTART'].dt == dt.timedelta(hours=1)


class TestIsAware:
    def test_naive(self):
        assert utils.is_aware(dt.datetime.now()) is False

    def test_berlin(self):
        assert utils.is_aware(BERLIN.localize(dt.datetime.now())) is True

    def test_bogota(self):
        assert utils.is_aware(BOGOTA.localize(dt.datetime.now())) is True

    def test_utc(self):
        assert utils.is_aware(pytz.UTC.localize(dt.datetime.now())) is True



import datetime as dt

import icalendar

from khal.ui.editor import RecurrenceEditor, StartEndEditor

from ..utils import BERLIN, LOCALE_BERLIN
from .canvas_render import CanvasTranslator

CONF = {'locale': LOCALE_BERLIN, 'keybindings': {}, 'view': {'monthdisplay': 'firstday'}}

START = BERLIN.localize(dt.datetime(2015, 4, 26, 22, 23))
END = BERLIN.localize(dt.datetime(2015, 4, 27, 23, 23))

palette = {
    'date header focused': 'blue',
    'date header': 'green',
    'default': 'black',
    'edit focused': 'red',
    'edit': 'blue',
}


def test_popup(monkeypatch):
    """making sure the popup calendar gets callend with the right inital value

    #405
    """
    class FakeCalendar:
        def store(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs

    fake = FakeCalendar()
    monkeypatch.setattr(
        'khal.ui.calendarwidget.CalendarWidget.__init__', fake.store)
    see = StartEndEditor(START, END, CONF)
    see.widgets.startdate.keypress((22, ), 'enter')
    assert fake.kwargs['initial'] == dt.date(2015, 4, 26)
    see.widgets.enddate.keypress((22, ), 'enter')
    assert fake.kwargs['initial'] == dt.date(2015, 4, 27)


def test_check_understood_rrule():
    assert RecurrenceEditor.check_understood_rrule(
        icalendar.vRecur.from_ical('FREQ=MONTHLY;BYDAY=1SU')
    )
    assert RecurrenceEditor.check_understood_rrule(
        icalendar.vRecur.from_ical('FREQ=MONTHLY;BYMONTHDAY=1')
    )
    assert RecurrenceEditor.check_understood_rrule(
        icalendar.vRecur.from_ical('FREQ=MONTHLY;BYDAY=TH;BYSETPOS=1')
    )
    assert RecurrenceEditor.check_understood_rrule(
        icalendar.vRecur.from_ical('FREQ=MONTHLY;BYDAY=TU,TH;BYSETPOS=1')
    )
    assert RecurrenceEditor.check_understood_rrule(
        icalendar.vRecur.from_ical('FREQ=MONTHLY;INTERVAL=2;BYDAY=MO,TU,WE,TH,FR,SA,SU;BYSETPOS=1')
    )
    assert RecurrenceEditor.check_understood_rrule(
        icalendar.vRecur.from_ical('FREQ=MONTHLY;INTERVAL=2;BYDAY=WE,SU,MO,TH,FR,TU,SA;BYSETPOS=1')
    )
    assert RecurrenceEditor.check_understood_rrule(
        icalendar.vRecur.from_ical('FREQ=MONTHLY;INTERVAL=2;BYDAY=WE,MO,TH,FR,TU,SA;BYSETPOS=1')
    )
    assert not RecurrenceEditor.check_understood_rrule(
        icalendar.vRecur.from_ical('FREQ=MONTHLY;BYDAY=-1SU')
    )
    assert not RecurrenceEditor.check_understood_rrule(
        icalendar.vRecur.from_ical('FREQ=MONTHLY;BYDAY=TH;BYMONTHDAY=1,2,3,4,5,6,7')
    )
    assert not RecurrenceEditor.check_understood_rrule(
        icalendar.vRecur.from_ical('FREQ=MONTHLY;BYDAY=TH;BYMONTHDAY=-1')
    )
    assert not RecurrenceEditor.check_understood_rrule(
        icalendar.vRecur.from_ical('FREQ=MONTHLY;BYDAY=TH;BYSETPOS=3')
    )


def test_editor():
    """test for the issue in #666"""
    editor = StartEndEditor(
        BERLIN.localize(dt.datetime(2017, 10, 2, 13)),
        BERLIN.localize(dt.datetime(2017, 10, 4, 18)),
        conf=CONF
    )
    assert editor.startdt == BERLIN.localize(dt.datetime(2017, 10, 2, 13))
    assert editor.enddt == BERLIN.localize(dt.datetime(2017, 10, 4, 18))
    assert editor.changed is False
    for _ in range(3):
        editor.keypress((10, ), 'tab')
    for _ in range(3):
        editor.keypress((10, ), 'shift tab')
    assert editor.startdt == BERLIN.localize(dt.datetime(2017, 10, 2, 13))
    assert editor.enddt == BERLIN.localize(dt.datetime(2017, 10, 4, 18))
    assert editor.changed is False


def test_convert_to_date():
    """test for the issue in #666"""
    editor = StartEndEditor(
        BERLIN.localize(dt.datetime(2017, 10, 2, 13)),
        BERLIN.localize(dt.datetime(2017, 10, 4, 18)),
        conf=CONF
    )
    canvas = editor.render((50, ), True)
    assert CanvasTranslator(canvas, palette).transform() == (
        '[ ] Allday\nFrom: \x1b[31m2.10.2017 \x1b[0m \x1b[34m13:00 \x1b[0m\n'
        'To:   \x1b[34m04.10.2017\x1b[0m \x1b[34m18:00 \x1b[0m\n'
    )

    assert editor.startdt == BERLIN.localize(dt.datetime(2017, 10, 2, 13))
    assert editor.enddt == BERLIN.localize(dt.datetime(2017, 10, 4, 18))
    assert editor.changed is False
    assert editor.allday is False

    # set to all day event
    editor.keypress((10, ), 'shift tab')
    editor.keypress((10, ), ' ')
    for _ in range(3):
        editor.keypress((10, ), 'tab')
    for _ in range(3):
        editor.keypress((10, ), 'shift tab')

    canvas = editor.render((50, ), True)
    assert CanvasTranslator(canvas, palette).transform() == (
        '[X] Allday\nFrom: \x1b[34m02.10.2017\x1b[0m  \n'
        'To:   \x1b[34m04.10.2017\x1b[0m  \n'
    )

    assert editor.changed is True
    assert editor.allday is True
    assert editor.startdt == dt.date(2017, 10, 2)
    assert editor.enddt == dt.date(2017, 10, 4)



import datetime as dt

import pytz
from packaging import version

from khal.khalendar.event import create_timezone

berlin = pytz.timezone('Europe/Berlin')
bogota = pytz.timezone('America/Bogota')

atime = dt.datetime(2014, 10, 28, 10, 10)
btime = dt.datetime(2016, 10, 28, 10, 10)


def test_berlin():
    vberlin_std = b'\r\n'.join(
        [b'BEGIN:STANDARD',
         b'DTSTART:20141026T020000',
         b'TZNAME:CET',
         b'TZOFFSETFROM:+0200',
         b'TZOFFSETTO:+0100',
         b'END:STANDARD',
         ])

    vberlin_dst = b'\r\n'.join(
        [b'BEGIN:DAYLIGHT',
         b'DTSTART:20150329T030000',
         b'TZNAME:CEST',
         b'TZOFFSETFROM:+0100',
         b'TZOFFSETTO:+0200',
         b'END:DAYLIGHT',
         ])

    vberlin = create_timezone(berlin, atime, atime).to_ical()
    assert b'TZID:Europe/Berlin' in vberlin
    assert vberlin_std in vberlin
    assert vberlin_dst in vberlin


def test_berlin_rdate():
    vberlin_std = b'\r\n'.join(
        [b'BEGIN:STANDARD',
         b'DTSTART:20141026T020000',
         b'RDATE:20151025T020000,20161030T020000',
         b'TZNAME:CET',
         b'TZOFFSETFROM:+0200',
         b'TZOFFSETTO:+0100',
         b'END:STANDARD',
         ])

    vberlin_dst = b'\r\n'.join(
        [b'BEGIN:DAYLIGHT',
         b'DTSTART:20150329T030000',
         b'RDATE:20160327T030000',
         b'TZNAME:CEST',
         b'TZOFFSETFROM:+0100',
         b'TZOFFSETTO:+0200',
         b'END:DAYLIGHT',
         ])

    vberlin = create_timezone(berlin, atime, btime).to_ical()
    assert b'TZID:Europe/Berlin' in vberlin
    assert vberlin_std in vberlin
    assert vberlin_dst in vberlin


def test_bogota():
    vbogota = [b'BEGIN:VTIMEZONE',
               b'TZID:America/Bogota',
               b'BEGIN:STANDARD',
               b'DTSTART:19930206T230000',
               b'TZNAME:COT',
               b'TZOFFSETFROM:-0400',
               b'TZOFFSETTO:-0500',
               b'END:STANDARD',
               b'END:VTIMEZONE',
               b'']
    if version.parse(pytz.__version__) > version.Version('2017.1'):
        vbogota[4] = b'TZNAME:-05'
        if version.parse(pytz.__version__) < version.Version('2022.7'):
            vbogota.insert(4, b'RDATE:20380118T221407')

    assert create_timezone(bogota, atime, atime).to_ical().split(b'\r\n') == vbogota



"""testing functions from the khal.utils"""
import datetime as dt

from click import style
from freezegun import freeze_time

from khal import utils


def test_relative_timedelta_str():
    with freeze_time('2016-9-19'):
        assert utils.relative_timedelta_str(dt.date(2016, 9, 24)) == '5 days from now'
        assert utils.relative_timedelta_str(dt.date(2016, 9, 29)) == '~1 week from now'
        assert utils.relative_timedelta_str(dt.date(2017, 9, 29)) == '~1 year from now'
        assert utils.relative_timedelta_str(dt.date(2016, 7, 29)) == '~7 weeks ago'


weekheader = """[1m    Mo Tu We Th Fr Sa Su   [0m"""
today_line = """[1mToday[0m[0m"""
calendarline = (
    "[1mNov [0m[1;33m31[0m [32m 1[0m [1;33m 2[0m "
    "[1;33m 3[0m [1;33m 4[0m [32m 5[0m [32m 6[0m"
)


def test_last_reset():
    assert utils.find_last_reset(weekheader) == (31, 35, '\x1b[0m')
    assert utils.find_last_reset(today_line) == (13, 17, '\x1b[0m')
    assert utils.find_last_reset(calendarline) == (99, 103, '\x1b[0m')
    assert utils.find_last_reset('Hello World') == (-2, -1, '')


def test_last_sgr():
    assert utils.find_last_sgr(weekheader) == (0, 4, '\x1b[1m')
    assert utils.find_last_sgr(today_line) == (0, 4, '\x1b[1m')
    assert utils.find_last_sgr(calendarline) == (92, 97, '\x1b[32m')
    assert utils.find_last_sgr('Hello World') == (-2, -1, '')


def test_find_unmatched_sgr():
    assert utils.find_unmatched_sgr(weekheader) is None
    assert utils.find_unmatched_sgr(today_line) is None
    assert utils.find_unmatched_sgr(calendarline) is None
    assert utils.find_unmatched_sgr('\x1b[31mHello World') == '\x1b[31m'
    assert utils.find_unmatched_sgr('\x1b[31mHello\x1b[0m \x1b[32mWorld') == '\x1b[32m'
    assert utils.find_unmatched_sgr('foo\x1b[1;31mbar') == '\x1b[1;31m'
    assert utils.find_unmatched_sgr('\x1b[0mfoo\x1b[1;31m') == '\x1b[1;31m'


def test_color_wrap():
    text = (
        "Lorem ipsum \x1b[31mdolor sit amet, consetetur sadipscing "
        "elitr, sed diam nonumy\x1b[0m eirmod tempor"
    )
    expected = [
        "Lorem ipsum \x1b[31mdolor sit amet,\x1b[0m",
        "\x1b[31mconsetetur sadipscing elitr, sed\x1b[0m",
        "\x1b[31mdiam nonumy\x1b[0m eirmod tempor",
    ]

    assert utils.color_wrap(text, 35) == expected


def test_color_wrap_256():
    text = (
        "\x1b[38;2;17;255;0mLorem ipsum dolor sit amet, consetetur sadipscing "
        "elitr, sed diam nonumy\x1b[0m"
    )
    expected = [
        "\x1b[38;2;17;255;0mLorem ipsum\x1b[0m",
        "\x1b[38;2;17;255;0mdolor sit amet, consetetur\x1b[0m",
        "\x1b[38;2;17;255;0msadipscing elitr, sed diam\x1b[0m",
        "\x1b[38;2;17;255;0mnonumy\x1b[0m"
    ]

    assert utils.color_wrap(text, 30) == expected


def test_color_wrap_multiple_colors_and_tabs():
    text = (
        "\x1b[31m14:00-14:50    AST-1002-102 INTRO AST II/STAR GALAX (R) Classes",
        "15:30-16:45    PHL-2000-104 PHILOSOPHY, SOCIETY & ETHICS (R) Classes",
        "\x1b[38;2;255;0m17:00-18:00    Pay Ticket Deadline Calendar",
        "09:30-10:45    PHL-1501-101 MIND, KNOWLEDGE & REALITY (R) Classes",
        "\x1b[38;2;255;0m11:00-14:00    Rivers Street (noodles and pizza) (R) Calendar",
    )
    expected = [
      '\x1b[31m14:00-14:50    AST-1002-102 INTRO AST II/STAR GALAX (R)\x1b[0m',
      '\x1b[31mClasses\x1b[0m',
      '15:30-16:45    PHL-2000-104 PHILOSOPHY, SOCIETY & ETHICS (R)',
      'Classes',
      '\x1b[38;2;255;0m17:00-18:00    Pay Ticket Deadline Calendar\x1b[0m',
      '09:30-10:45    PHL-1501-101 MIND, KNOWLEDGE & REALITY (R)',
      'Classes',
      '\x1b[38;2;255;0m11:00-14:00    Rivers Street (noodles and\x1b[0m',
      '\x1b[38;2;255;0mpizza) (R) Calendar\x1b[0m'
    ]
    actual = []
    for line in text:
        actual += utils.color_wrap(line, 60)
    assert actual == expected


def test_get_weekday_occurrence():
    assert utils.get_weekday_occurrence(dt.datetime(2017, 3, 1)) == (2, 1)
    assert utils.get_weekday_occurrence(dt.datetime(2017, 3, 2)) == (3, 1)
    assert utils.get_weekday_occurrence(dt.datetime(2017, 3, 3)) == (4, 1)
    assert utils.get_weekday_occurrence(dt.datetime(2017, 3, 4)) == (5, 1)
    assert utils.get_weekday_occurrence(dt.datetime(2017, 3, 5)) == (6, 1)
    assert utils.get_weekday_occurrence(dt.datetime(2017, 3, 6)) == (0, 1)
    assert utils.get_weekday_occurrence(dt.datetime(2017, 3, 7)) == (1, 1)
    assert utils.get_weekday_occurrence(dt.datetime(2017, 3, 8)) == (2, 2)
    assert utils.get_weekday_occurrence(dt.datetime(2017, 3, 9)) == (3, 2)
    assert utils.get_weekday_occurrence(dt.datetime(2017, 3, 10)) == (4, 2)

    assert utils.get_weekday_occurrence(dt.datetime(2017, 3, 31)) == (4, 5)

    assert utils.get_weekday_occurrence(dt.date(2017, 5, 1)) == (0, 1)
    assert utils.get_weekday_occurrence(dt.date(2017, 5, 7)) == (6, 1)
    assert utils.get_weekday_occurrence(dt.date(2017, 5, 8)) == (0, 2)
    assert utils.get_weekday_occurrence(dt.date(2017, 5, 28)) == (6, 4)
    assert utils.get_weekday_occurrence(dt.date(2017, 5, 29)) == (0, 5)


def test_human_formatter_width():
    formatter = utils.human_formatter('{red}{title}', width=10)
    output = formatter({'title': 'morethan10characters', 'red': style('', reset=False, fg='red')})
    assert output.startswith('\x1b[31mmoret\x1b[0m')
